from abc import abstractmethod
from typing import Any, Dict, Iterator, Optional, Tuple, Union

import torch

import pystiche
from pystiche.enc import Encoder

__all__ = [
    "Operator",
    "RegularizationOperator",
    "ComparisonOperator",
    "PixelOperator",
    "EncodingOperator",
    "PixelRegularizationOperator",
    "EncodingRegularizationOperator",
    "PixelComparisonOperator",
    "EncodingComparisonOperator",
]


class Operator(pystiche.Module):
    r"""Abstract base class for all operators. If called, invokes
    :meth:`pystiche.ops.Operator.process_input_image` and applies ``score_weight`` to
    the result.

    Args:
        score_weight: Score weight of the operator. Defaults to ``1.0``.
    """

    def __init__(self, score_weight: float = 1e0,) -> None:
        super().__init__()
        self.score_weight = score_weight

    input_guide: torch.Tensor

    def set_input_guide(self, guide: torch.Tensor) -> None:
        r"""Set input guide.

        Args:
            guide: Input guide of shape :math:`1 \times 1 \times H \times W`.
        """
        self.register_buffer("input_guide", guide, persistent=False)

    @property
    def has_input_guide(self) -> bool:
        return "input_guide" in self._buffers

    @staticmethod
    def apply_guide(image: torch.Tensor, guide: torch.Tensor) -> torch.Tensor:
        r"""Apply a guide to an image.

        Args:
            image: Image of shape :math:`B \times C \times H \times W`.
            guide: Guide of shape :math:`1 \times 1 \times H \times W`.
        """
        return image * guide

    def forward(
        self, input_image: torch.Tensor
    ) -> Union[torch.Tensor, pystiche.LossDict]:
        return self.process_input_image(input_image) * self.score_weight

    @abstractmethod
    def process_input_image(
        self, image: torch.Tensor
    ) -> Union[torch.Tensor, pystiche.LossDict]:
        r"""Defines the computation performed with every call.

        Args:
            image: Input image of shape :math:`B \times C \times H \times W`.
        """
        pass

    def named_operators(
        self, recurse: bool = False,
    ) -> Iterator[Tuple[str, "Operator"]]:
        if recurse:
            iterator = self.named_modules()
            # First module is always self so dismiss that
            next(iterator)
        else:
            iterator = self.named_children()
        for name, child in iterator:
            if isinstance(child, Operator):
                yield name, child

    def operators(self, recurse: bool = False) -> Iterator["Operator"]:
        for _, op in self.named_operators(recurse=recurse):
            yield op

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["score_weight"] = f"{self.score_weight:g}"
        return dct


class RegularizationOperator(Operator):
    r"""Abstract base class for all regularization operators."""

    @abstractmethod
    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        pass


class ComparisonOperator(Operator):
    r"""Abstract base class for all comparison operators."""
    target_guide: torch.Tensor
    target_image: torch.Tensor
    target_repr: torch.Tensor
    ctx: Optional[torch.Tensor]

    @abstractmethod
    def set_target_guide(self, guide: torch.Tensor, recalc_repr: bool = True) -> None:
        r"""Set a target guide and optionally recalculate the target representation.

        Args:
            guide: Input guide of shape :math:`1 \times 1 \times H \times W`.
            recalc_repr: If ``True``, recalculates :meth:`.target_enc_to_repr`.
                Defaults to ``True``.

                .. note::

                    Set this to ``False`` if the shape of ``guide`` and the shape of a
                    previously set ``target_image`` do not match.
        """
        pass

    @property
    def has_target_guide(self) -> bool:
        return "target_guide" in self._buffers

    @abstractmethod
    def set_target_image(self, image: torch.Tensor) -> None:
        r"""Set a target image and calculate its representation.

        Args:
            image: Target image of shape :math:`B \times C \times H \times W`.
        """
        pass

    @property
    def has_target_image(self) -> bool:
        return "target_image" in self._buffers

    @abstractmethod
    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        pass

    @staticmethod
    def _match_batch_sizes(target: torch.Tensor, input: torch.Tensor) -> torch.Tensor:
        target_batch_size = target.size()[0]
        input_batch_size = input.size()[0]

        if target_batch_size == input_batch_size:
            return target

        if target_batch_size != 1:
            raise RuntimeError(
                f"If the batch size of the target != 1, "
                f"it has to match the batch size of the input. "
                f"Got {target_batch_size} != {input_batch_size}"
            )

        with torch.no_grad():
            return target.repeat(input_batch_size, *[1] * (target.dim() - 1))


class PixelOperator(Operator):
    r"""Abstract base class for all operators working in the pixel space."""

    @abstractmethod
    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        pass


class EncodingOperator(Operator):
    r"""Abstract base class for all operators working in an encoded space."""

    encoder: Encoder

    @abstractmethod
    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        pass

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["encoder"] = self.encoder
        return dct

    def __repr__(self) -> str:
        return self._build_repr(
            named_children=[
                (name, child)
                for name, child in self.named_children()
                if child is not self.encoder
            ]
        )


class PixelRegularizationOperator(PixelOperator, RegularizationOperator):
    r"""Abstract base class for all regularization operators working in the pixel space.

    Args:
        score_weight: Score weight of the operator. Defaults to ``1.0``.
    """

    def __init__(self, score_weight: float = 1.0):
        super().__init__(score_weight=score_weight,)

    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        if self.has_input_guide:
            image = self.apply_guide(image, self.input_guide)
        input_repr = self.input_image_to_repr(image)
        return self.calculate_score(input_repr)

    @abstractmethod
    def input_image_to_repr(self, image: torch.Tensor) -> torch.Tensor:
        r"""Calculate the input representation from the input image.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            image: Input image of shape :math:`B \times C \times H \times W`.
        """
        pass

    @abstractmethod
    def calculate_score(self, input_repr: torch.Tensor) -> torch.Tensor:
        r"""Calculate the operator score from the input representation.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            input_repr: Input representation.
        """
        pass


class EncodingRegularizationOperator(EncodingOperator, RegularizationOperator):
    r"""Abstract base class for all regularization operators working in an encoded
    space.

    Args:
        encoder: Encoder that is used to encode the target and input images.
        score_weight: Score weight of the operator. Defaults to ``1.0``.
    """

    def __init__(self, encoder: Encoder, score_weight: float = 1.0):
        super().__init__(score_weight=score_weight)
        self.encoder = encoder

    def set_input_guide(self, guide: torch.Tensor) -> None:
        super().set_input_guide(guide)
        with torch.no_grad():
            enc_guide = self.encoder.propagate_guide(guide)
        self.register_buffer("input_enc_guide", enc_guide, persistent=False)

    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        input_repr = self.input_image_to_repr(image)
        return self.calculate_score(input_repr)

    def input_image_to_repr(self, image: torch.Tensor) -> torch.Tensor:
        enc = self.encoder(image)
        if self.has_input_guide:
            enc = self.apply_guide(enc, self.input_guide)
        return self.input_enc_to_repr(enc)

    @abstractmethod
    def input_enc_to_repr(self, enc: torch.Tensor) -> torch.Tensor:
        r"""Calculate the input representation from the encoded input image.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            enc: Encoded input image of shape :math:`B \times C \times H \times W`.
        """
        pass

    @abstractmethod
    def calculate_score(self, input_repr: torch.Tensor) -> torch.Tensor:
        r"""Calculate the operator score from the input representation.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            input_repr: Input representation.
        """
        pass


class PixelComparisonOperator(PixelOperator, ComparisonOperator):
    r"""Abstract base class for all comparison operators working in the pixel space.

    Args:
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Raises:
        RuntimeError: If called without setting a target image first.
    """

    def __init__(self, score_weight: float = 1e0):
        super().__init__(score_weight=score_weight)

    def set_target_guide(self, guide: torch.Tensor, recalc_repr: bool = True) -> None:
        self.register_buffer("target_guide", guide, persistent=False)
        if recalc_repr and self.has_target_image:
            self.set_target_image(self.target_image)

    def set_target_image(self, image: torch.Tensor) -> None:
        self.register_buffer("target_image", image, persistent=False)
        with torch.no_grad():
            if self.has_target_guide:
                image = self.apply_guide(image, self.target_guide)
            repr, ctx = self.target_image_to_repr(image)
        self.register_buffer("target_repr", repr, persistent=False)
        if ctx is not None:
            self.register_buffer("ctx", ctx, persistent=False)
        else:
            self.ctx = None

    @abstractmethod
    def target_image_to_repr(
        self, image: torch.Tensor
    ) -> Tuple[
        torch.Tensor, Optional[torch.Tensor],
    ]:
        r"""Calculate the target representation and context information from the
        target image.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            image: Target image of shape :math:`B \times C \times H \times W`.
        """
        pass

    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        if not self.has_target_image:
            msg = "Cannot process an input image before a target image has been set."
            raise RuntimeError(msg)
        target_repr, ctx = self.target_repr, self.ctx

        if self.has_input_guide:
            image = self.apply_guide(image, self.input_guide)
        input_repr = self.input_image_to_repr(image, ctx)

        target_repr = self._match_batch_sizes(target_repr, input_repr)
        return self.calculate_score(input_repr, target_repr, ctx)

    @abstractmethod
    def input_image_to_repr(
        self, image: torch.Tensor, ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        r"""Calculate the input representation from the input image optionally using
        the context information.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            image: Input image of shape :math:`B \times C \times H \times W`.
            ctx: Optional context information.
        """
        pass

    @abstractmethod
    def calculate_score(
        self,
        input_repr: torch.Tensor,
        target_repr: torch.Tensor,
        ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        r"""Calculate the operator score from the input and target representation
        optionally using the context information.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            input_repr: Input representation.
            target_repr: Target representation.
            ctx: Optional context information.
        """
        pass


class EncodingComparisonOperator(EncodingOperator, ComparisonOperator):
    r"""Abstract base class for all comparison operators working in an encoded space.

    Args:
        encoder: Encoder that is used to encode the target and input images.
        score_weight: Score weight of the operator. Defaults to ``1.0``.

    Raises:
        RuntimeError: If called without setting a target image first.
    """
    target_enc_guide: torch.Tensor
    input_enc_guide: torch.Tensor

    def __init__(self, encoder: Encoder, score_weight: float = 1.0):
        super().__init__(score_weight=score_weight)
        self.encoder = encoder

    def set_target_guide(self, guide: torch.Tensor, recalc_repr: bool = True) -> None:

        with torch.no_grad():
            enc_guide = self.encoder.propagate_guide(guide)
        self.register_buffer("target_guide", guide, persistent=False)
        self.register_buffer("target_enc_guide", enc_guide, persistent=False)
        if recalc_repr and self.has_target_image:
            self.set_target_image(self.target_image)

    def set_target_image(self, image: torch.Tensor) -> None:

        with torch.no_grad():
            repr, ctx = self.target_image_to_repr(image)
        self.register_buffer("target_image", image, persistent=False)
        self.register_buffer("target_repr", repr, persistent=False)
        if ctx is not None:
            self.register_buffer("ctx", ctx, persistent=False)
        else:
            self.ctx = None

    def target_image_to_repr(
        self, image: torch.Tensor
    ) -> Tuple[
        torch.Tensor, Optional[torch.Tensor],
    ]:
        enc = self.encoder(image)
        if self.has_target_guide:
            enc = self.apply_guide(enc, self.target_enc_guide)
        return self.target_enc_to_repr(enc)

    @abstractmethod
    def target_enc_to_repr(
        self, enc: torch.Tensor
    ) -> Tuple[
        torch.Tensor, Optional[torch.Tensor],
    ]:
        r"""Calculate the target representation and context information from the
        encoded target image.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            enc: Encoded target image of shape :math:`B \times C \times H \times W`.
        """

    def set_input_guide(self, guide: torch.Tensor) -> None:
        with torch.no_grad():
            enc_guide = self.encoder.propagate_guide(guide)
        self.register_buffer("input_guide", guide, persistent=False)
        self.register_buffer("input_enc_guide", enc_guide, persistent=False)

    def process_input_image(self, image: torch.Tensor) -> torch.Tensor:
        if not self.has_target_image:
            msg = "Cannot process an input image before a target image has been set."
            raise RuntimeError(msg)
        target_repr, ctx = self.target_repr, self.ctx
        input_repr = self.input_image_to_repr(image, ctx)

        target_repr = self._match_batch_sizes(target_repr, input_repr)
        return self.calculate_score(input_repr, target_repr, ctx)

    def input_image_to_repr(
        self, image: torch.Tensor, ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        enc = self.encoder(image)
        if self.has_input_guide:
            enc = self.apply_guide(enc, self.input_enc_guide)
        return self.input_enc_to_repr(enc, ctx)

    @abstractmethod
    def input_enc_to_repr(
        self, enc: torch.Tensor, ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        r"""Calculate the input representation from the encoded input image optionally
        using the context information.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            enc: Encoded input image of shape :math:`B \times C \times H \times W`.
            ctx: Optional context information.
        """
        pass

    @abstractmethod
    def calculate_score(
        self,
        input_repr: torch.Tensor,
        target_repr: torch.Tensor,
        ctx: Optional[torch.Tensor],
    ) -> torch.Tensor:
        r"""Calculate the operator score from the input and target representation
        optionally using the context information.

        .. note::

            This method has to be overwritten in every subclass.

        Args:
            input_repr: Input representation.
            target_repr: Target representation.
            ctx: Optional context information.
        """
        pass



from typing import Any, Dict, Iterator, Optional

import torch

from pystiche import ComplexObject
from pystiche.misc import suppress_warnings, verify_str_arg

with suppress_warnings():
    from pystiche.image.transforms.functional import resize

__all__ = ["PyramidLevel"]


class PyramidLevel(ComplexObject):
    r"""Level with an :class:`pystiche.pyramid.ImagePyramid`. If iterated on, yields
    the step beginning at 1 and ending in ``num_steps``.

    Args:
        edge_size: Edge size.
        num_steps: Number of steps.
        edge: Corresponding edge to the edge size. Can be ``"short"`` or
            ``"long"``.
    """

    def __init__(self, edge_size: int, num_steps: int, edge: str) -> None:
        self.edge_size = edge_size
        self.num_steps = num_steps
        self.edge = verify_str_arg(edge, "edge", ("short", "long"))

    def _resize(
        self,
        image: torch.Tensor,
        aspect_ratio: Optional[float],
        interpolation_mode: str,
    ) -> torch.Tensor:

        with torch.no_grad():
            image = resize(
                image,
                self.edge_size,
                edge=self.edge,
                aspect_ratio=aspect_ratio,
                interpolation_mode=interpolation_mode,
            )
        return image.detach()

    def resize_image(
        self,
        image: torch.Tensor,
        aspect_ratio: Optional[float] = None,
        interpolation_mode: str = "bilinear",
    ) -> torch.Tensor:
        r"""Resize an image to the ``edge_size`` on the corresponding ``edge`` of the
        :class:`PyramidLevel`.

        Args:
            image: Image to be resized.
            aspect_ratio: Optional aspect ratio of the output. If ``None``, the aspect
                ratio of ``image`` is used. Defaults to ``None``.
            interpolation_mode: Interpolation mode used to resize ``image``. Defaults
                to ``"bilinear"``.

        .. warning::

            The resizing is performed without gradient calculation. Do **not** use this
            if the image needs a gradient.
        """
        return self._resize(image, aspect_ratio, interpolation_mode)

    def resize_guide(
        self,
        guide: torch.Tensor,
        aspect_ratio: Optional[float] = None,
        interpolation_mode: str = "nearest",
    ) -> torch.Tensor:
        r"""Resize a guide to the ``edge_size`` on the corresponding ``edge`` of the
        :class:`PyramidLevel`.

        Args:
            guide: Guide to be resized.
            aspect_ratio: Optional aspect ratio of the output. If ``None``, the aspect
                ratio of ``guide`` is used. Defaults to ``None``.
            interpolation_mode: Interpolation mode used to resize ``image``. Defaults
                to ``"nearest"``.
        """
        return self._resize(guide, aspect_ratio, interpolation_mode)

    def __iter__(self) -> Iterator[int]:
        yield from range(1, self.num_steps + 1)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["edge_size"] = self.edge_size
        dct["num_steps"] = self.num_steps
        dct["edge"] = self.edge
        return dct



import warnings
from abc import ABC, abstractmethod
from collections import OrderedDict, deque
from datetime import datetime, timedelta
from typing import Any, Callable, Optional, Sequence, Union, cast

import torch

import pystiche
from pystiche.misc import build_deprecation_message, build_fmtstr

__all__ = [
    "Meter",
    "FloatMeter",
    "AverageMeter",
    "LossMeter",
    "ETAMeter",
    "ProgressMeter",
]


def _deprecation_warning() -> None:
    msg = build_deprecation_message(
        "Using any functionality from pystiche.optim.meter",
        "0.7.0",
        info="See https://github.com/pmeier/pystiche/issues/434 for details.",
    )
    warnings.warn(msg, UserWarning)


class Meter(ABC):
    def __init__(self, name: Optional[str] = None):
        _deprecation_warning()
        self.name = name

    @abstractmethod
    def reset(self) -> None:
        pass

    update: Callable[..., None]

    @abstractmethod
    def __str__(self) -> str:
        pass


class FloatMeter(Meter):
    count: int
    last_val: float
    global_sum: float
    global_min: float
    global_max: float
    window: deque

    def __init__(self, name: str, window_size: int = 50) -> None:
        super().__init__(name)
        self.window_size = window_size
        self.reset()

    def reset(self) -> None:
        self.count = 0
        self.last_val = 0.0
        self.global_sum = 0.0
        self.global_min = float("inf")
        self.global_max = -float("inf")
        self.window = deque(maxlen=self.window_size)

    def update(self, vals: Union[Sequence[float], float]) -> None:
        if isinstance(vals, float):
            vals = (vals,)

        self.count += len(vals)
        self.last_val = vals[-1]
        self.global_sum += sum(vals)
        self.global_min = min(*vals, self.global_min)
        self.global_max = max(*vals, self.global_max)
        self.window.extend(vals)

    @property
    def global_avg(self) -> float:
        if self.count == 0:
            msg = "global_avg can only be calculated after the first update."
            raise RuntimeError(msg)
        return self.global_sum / self.count

    @property
    def local_avg(self) -> float:
        vals = tuple(self.window)
        if not vals:
            msg = "local_avg can only be calculated after the first update."
            raise RuntimeError(msg)
        return sum(vals) / len(vals)

    @abstractmethod
    def __str__(self) -> str:
        pass


class AverageMeter(FloatMeter):
    def __init__(
        self,
        name: str,
        window_size: int = 50,
        show_local_avg: bool = True,
        fmt: str = "{:f}",
    ):
        super().__init__(name=name, window_size=window_size)
        self.show_local_avg = show_local_avg
        self.fmt = fmt

    def update(self, vals: Union[Sequence[float], float, torch.Tensor]) -> None:
        if isinstance(vals, torch.Tensor):
            vals = vals.detach().cpu().flatten().tolist()

        super().update(vals)

    def __str__(self) -> str:
        def format(val: float) -> str:
            return self.fmt.format(val)

        if self.count > 0:
            val = format(self.last_val)
            avg = format(self.local_avg if self.show_local_avg else self.global_avg)
            info = f"{val} ({avg})"
        else:
            info = "N/A"
        return f"{self.name} {info}"


class LossMeter(AverageMeter):
    def __init__(self, name: str = "loss", fmt: str = "{:.3e}", **kwargs: Any) -> None:
        super().__init__(name, fmt=fmt, **kwargs)

    def update(
        self, val: Union[Sequence[float], float, torch.Tensor, pystiche.LossDict],
    ) -> None:
        if isinstance(val, pystiche.LossDict):
            val = float(val)
        super().update(val)


class ETAMeter(FloatMeter):
    last_time: Optional[float]

    def __init__(
        self,
        total_count: int,
        name: str = "ETA",
        window_size: int = 500,
        show_local_eta: bool = True,
        fmt: str = "%d.%m.%Y %H:%M",
    ):
        super().__init__(name=name, window_size=window_size)
        self.total_count = total_count
        self.show_local_eta = show_local_eta
        self.fmt = fmt

        self.reset()

    def reset(self) -> None:
        super().reset()
        self.last_time = None

    # This violates the Liskov Substitution Principle. We ignore this for now since
    # this complete module should be refactored in the future.
    def update(self, time: float) -> None:  # type: ignore[override]
        if self.last_time is None:
            self.last_time = time
            return

        time_diff = time - self.last_time
        super().update(time_diff)
        self.last_time = time

    def calculate_eta(self, time_diff: float) -> datetime:
        count_diff = max(self.total_count - self.count, 0)
        return datetime.now() + count_diff * timedelta(seconds=time_diff)

    @property
    def global_eta(self) -> datetime:
        return self.calculate_eta(self.global_avg)

    @property
    def local_eta(self) -> datetime:
        return self.calculate_eta(self.local_avg)

    def __str__(self) -> str:
        if self.count > 0:
            eta = self.local_eta if self.show_local_eta else self.global_eta
            info = eta.strftime(self.fmt)
        else:
            info = "N/A"
        return f"{self.name} {info}"


class ProgressMeter(Meter):
    count: int
    progress_fmt: str

    def __init__(
        self,
        total_count: int,
        *meters: Meter,
        name: Optional[str] = None,
        build_progress_fmt: Optional[Callable[[int], str]] = None,
    ) -> None:
        super().__init__(name=name)
        self.meters = OrderedDict([(meter.name, meter) for meter in meters])

        if build_progress_fmt is None:

            def build_progress_fmt(total_count: int) -> str:
                count_fmt = build_fmtstr(field_len=len(str(total_count)), type="d")
                return f"[{count_fmt}/{total_count:d}]"

            build_progress_fmt = cast(Callable[[int], str], build_progress_fmt)

        self.build_progress_fmt = build_progress_fmt

        self.reset(total_count=total_count)

    def reset(self, total_count: Optional[int] = None,) -> None:
        self.count = 0

        if total_count is not None:
            self.progress_fmt = self.build_progress_fmt(total_count)

        for meter in self.meters.values():
            meter.reset()

    def update(self, **kwargs: Any) -> None:
        self.count += 1

        for name, vals in kwargs.items():
            self.meters[name].update(vals)

    @property
    def progress(self) -> str:
        return self.progress_fmt.format(self.count)

    def __str__(self) -> str:
        parts = []

        if self.name is not None:
            parts.append(self.name)

        parts.append(self.progress)

        if self.meters:
            parts.extend([str(meter) for meter in self.meters.values()])

        return "\t".join(parts)



import os
import warnings
from typing import Any, Callable, Iterator, List, Optional, Tuple, cast

import torch
from torch import nn
from torch.utils.data import Dataset
from torchvision.datasets import VisionDataset
from torchvision.datasets.folder import is_image_file

from pystiche.image import read_image
from pystiche.misc import build_deprecation_message

__all__ = [
    "Unsupervised",
    "ImageFolderDataset",
]


class Unsupervised:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        msg = build_deprecation_message(
            "The class 'Unsupervised'",
            "0.7",
            info="See https://github.com/pmeier/pystiche/issues/458 for details",
        )
        warnings.warn(msg)
        if not isinstance(self, VisionDataset):
            raise RuntimeError
        # This should be used with double inheritance with a
        # torchvision.datasets.VisionDataset
        super().__init__(*args, **kwargs)  # type: ignore[call-arg]

    def __getitem__(self, index: int) -> Any:
        # This should be used with double inheritance with a
        # torchvision.datasets.VisionDataset
        return super().__getitem__(index)[0]  # type: ignore[misc]


def walkupto(
    top: str, depth: Optional[int] = None, **kwargs: Any
) -> Iterator[Tuple[str, List[str], List[str]]]:
    if depth is None:
        yield from os.walk(top, **kwargs)
        return

    base_depth = top.count(os.sep)
    for root, dirs, files in os.walk(top, **kwargs):
        if root.count(os.sep) <= base_depth + depth:
            yield root, dirs, files
            # FIXME: stop walking directories if top directory is already to deep


class ImageFolderDataset(Dataset):
    def __init__(
        self,
        root: str,
        transform: Optional[nn.Module] = None,
        depth: Optional[int] = None,
        importer: Optional[Callable[[str], Any]] = None,
    ):
        self.root = os.path.abspath(os.path.expanduser(root))
        self.image_files = self._collect_image_files(depth)
        self.transform = transform

        if importer is None:

            def importer(file: str) -> torch.Tensor:
                return read_image(file, make_batched=False)

        self.importer = cast(Callable[[str], Any], importer)

    def _collect_image_files(self, depth: Optional[int]) -> Tuple[str, ...]:
        image_files = tuple(
            os.path.join(root, file)
            for root, _, files in walkupto(self.root, depth=depth)
            for file in files
            if is_image_file(file)
        )

        if not image_files:
            msg = f"The directory {self.root} does not contain any image files"
            if depth is not None:
                msg += f" up to a depth of {depth}"
            msg += "."
            raise RuntimeError(msg)

        return image_files

    def __len__(self) -> int:
        return len(self.image_files)

    def __getitem__(self, idx: int) -> Any:
        file = self.image_files[idx]
        image = self.importer(file)

        if self.transform:
            image = self.transform(image)

        return image



from abc import abstractmethod
from typing import Sequence

import torch
from torch import nn

import pystiche

from .guides import propagate_guide

__all__ = ["Encoder", "SequentialEncoder"]


class Encoder(pystiche.Module):
    r"""ABC for all encoders. Invokes :meth:`Encoder.forward` if called."""

    @abstractmethod
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        r"""Encodes the given input.

        .. note::

            This method has to be overwritten in every subclass.
        """
        pass

    @abstractmethod
    def propagate_guide(self, guide: torch.Tensor) -> torch.Tensor:
        r"""Encodes the given guide.

        .. note::

            This method has to be overwritten in every subclass.
        """
        pass


class SequentialEncoder(Encoder):
    r"""Encoder that operates in sequential manner. Invokes :meth:`Encoder.forward`
    if called.

    Args:
        modules: Sequential modules.
    """

    def __init__(self, modules: Sequence[nn.Module]) -> None:
        super().__init__(indexed_children=modules)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        for module in self.children():
            input = module(input)
        return input

    def propagate_guide(self, guide: torch.Tensor) -> torch.Tensor:
        for module in self.children():
            guide = propagate_guide(module, guide)
        return guide



from typing import Optional, Tuple, Union, cast

import torch

from pystiche.image.utils import (
    edge_to_image_size,
    extract_aspect_ratio,
    extract_image_size,
    force_batched_image,
    is_edge_size,
    is_image_size,
)
from pystiche.misc import to_2d_arg

from ._align_corners import interpolate

__all__ = ["resize", "rescale"]


@force_batched_image
def resize(
    image: torch.Tensor,
    size: Union[int, Tuple[int, int]],
    edge: str = "short",
    aspect_ratio: Optional[float] = None,
    interpolation_mode: str = "bilinear",
) -> torch.Tensor:
    r"""Resize an image as specified. See :class:`~pystiche.image.transforms.Resize`
    for details.
    """
    if aspect_ratio is None:
        aspect_ratio = extract_aspect_ratio(image)

    if is_image_size(size):
        image_size = size
    elif is_edge_size(size):
        image_size = edge_to_image_size(cast(int, size), aspect_ratio, edge)
    else:
        msg = (
            f"size can either be an edge size (int) or an image size "
            f"(Tuple[int, int]), but got {type(size)}."
        )
        raise TypeError(msg)

    return interpolate(
        image, size=image_size, scale_factor=None, mode=interpolation_mode,
    )


def rescale(
    image: torch.Tensor,
    factor: Union[float, Tuple[float, float]],
    interpolation_mode: str = "bilinear",
) -> torch.Tensor:
    height, width = extract_image_size(image)
    height_factor, width_factor = to_2d_arg(factor)
    height = round(height * height_factor)
    width = round(width * width_factor)
    image_size = (height, width)

    return cast(
        torch.Tensor, resize(image, image_size, interpolation_mode=interpolation_mode)
    )



from typing import Tuple, Union, cast

import torch

from pystiche.image.utils import (
    extract_image_size,
    force_batched_image,
    is_edge_size,
    is_image_size,
)
from pystiche.misc import verify_str_arg

__all__ = [
    "crop",
    "top_left_crop",
    "bottom_left_crop",
    "top_right_crop",
    "bottom_right_crop",
    "center_crop",
]


def _parse_size(size: Union[Tuple[int, int], int]) -> Tuple[int, int]:
    if is_image_size(size):
        return cast(Tuple[int, int], size)
    elif is_edge_size(size):
        edge_size = cast(int, size)
        return edge_size, edge_size
    else:
        msg = (
            f"size can either be an edge size (int) or an image size "
            f"(Tuple[int, int]), but got {type(size)}."
        )
        raise TypeError(msg)


@force_batched_image
def crop(
    x: torch.Tensor,
    origin: Tuple[int, int],
    size: Union[Tuple[int, int], int],
    vert_anchor: str = "top",
    horz_anchor: str = "left",
) -> torch.Tensor:
    verify_str_arg(vert_anchor, "vert_anchor", ("top", "bottom"))
    verify_str_arg(horz_anchor, "horz_anchor", ("left", "right"))

    vert_origin, horz_origin = origin
    height, width = _parse_size(size)

    def create_vert_slice() -> slice:
        if vert_anchor == "top":
            return slice(vert_origin, vert_origin + height)
        else:  # vert_anchor == "bottom"
            return slice(vert_origin - height, vert_origin)

    def create_horz_slice() -> slice:
        if horz_anchor == "left":
            return slice(horz_origin, horz_origin + width)
        else:  # horz_anchor == "right"
            return slice(horz_origin - width, horz_origin)

    vert_slice = create_vert_slice()
    horz_slice = create_horz_slice()
    return x[:, :, vert_slice, horz_slice]


def top_left_crop(x: torch.Tensor, size: Union[Tuple[int, int], int]) -> torch.Tensor:
    origin = (0, 0)
    return cast(
        torch.Tensor, crop(x, origin, size, vert_anchor="top", horz_anchor="left")
    )


def bottom_left_crop(
    x: torch.Tensor, size: Union[Tuple[int, int], int]
) -> torch.Tensor:
    height, _ = extract_image_size(x)
    origin = (height, 0)
    return cast(
        torch.Tensor, crop(x, origin, size, vert_anchor="bottom", horz_anchor="left")
    )


def top_right_crop(x: torch.Tensor, size: Union[Tuple[int, int], int]) -> torch.Tensor:
    _, width = extract_image_size(x)
    origin = (0, width)
    return cast(
        torch.Tensor, crop(x, origin, size, vert_anchor="top", horz_anchor="right")
    )


def bottom_right_crop(
    x: torch.Tensor, size: Union[Tuple[int, int], int]
) -> torch.Tensor:
    origin = extract_image_size(x)
    return cast(
        torch.Tensor, crop(x, origin, size, vert_anchor="bottom", horz_anchor="right")
    )


def center_crop(x: torch.Tensor, size: Union[Tuple[int, int], int]) -> torch.Tensor:
    image_size = extract_image_size(x)
    size = _parse_size(size)
    vert_origin = (image_size[0] - size[0]) // 2
    horz_origin = (image_size[1] - size[1]) // 2
    return cast(torch.Tensor, crop(x, (vert_origin, horz_origin), size))



from typing import Dict, Iterable, Optional, Sequence, Tuple, cast

import torch
from torch import nn

from ._objects import ComplexObject

__all__ = ["Module", "SequentialModule"]


class Module(nn.Module, ComplexObject):
    r""":class:`torch.nn.Module` with the enhanced representation options of
    :class:`pystiche.ComplexObject`.

    Args:
        named_children: Optional children modules that are added by their names.
        indexed_children: Optional children modules that are added with successive
            indices as names.

    .. note::

        ``named_children`` and ``indexed_children`` are mutually exclusive parameters.
    """
    _buffers: Dict[str, torch.Tensor]
    _modules: Dict[str, nn.Module]

    def __init__(
        self,
        named_children: Optional[Iterable[Tuple[str, nn.Module]]] = None,
        indexed_children: Optional[Sequence[nn.Module]] = None,
    ):
        super().__init__()
        if named_children is not None and indexed_children is not None:
            msg = (
                "named_children and indexed_children "
                "are mutually exclusive parameters."
            )
            raise RuntimeError(msg)
        elif named_children is not None:
            self.add_named_modules(named_children)
        elif indexed_children is not None:
            self.add_indexed_modules(indexed_children)

    def add_named_modules(self, modules: Iterable[Tuple[str, nn.Module]]) -> None:
        for name, module in modules:
            self.add_module(name, module)

    def add_indexed_modules(self, modules: Sequence[nn.Module]) -> None:
        self.add_named_modules(
            [(str(idx), module) for idx, module in enumerate(modules)]
        )

    def __repr__(self) -> str:
        return ComplexObject.__repr__(self)

    def torch_repr(self) -> str:
        r"""
        Returns:
            Native torch representation.
        """
        return cast(str, nn.Module.__repr__(self))

    def extra_repr(self) -> str:
        return ", ".join([f"{key}={value}" for key, value in self.properties().items()])


class SequentialModule(Module):
    def __init__(self, *modules: nn.Module):
        super().__init__(indexed_children=modules)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for module in self.children():
            x = module(x)
        return x



import functools
from typing import Any, Callable, Optional

import torch

from .core import ComposedTransform
from .misc import (
    Denormalize,
    FloatToUint8Range,
    Normalize,
    ReverseChannelOrder,
    Uint8ToFloatRange,
)

TORCH_MEAN = (0.485, 0.456, 0.406)
TORCH_STD = (0.229, 0.224, 0.225)

CAFFE_MEAN = (0.485, 0.458, 0.408)
CAFFE_STD = (1.0, 1.0, 1.0)

__all__ = [
    "TorchPreprocessing",
    "TorchPostprocessing",
    "torch_processing",
    "CaffePreprocessing",
    "CaffePostprocessing",
    "caffe_processing",
]


def _processing(
    preprocessor: Optional[Callable] = None, postprocessor: Optional[Callable] = None
) -> Callable[[Callable], Callable]:
    def outer_wrapper(fn: Callable) -> Callable:
        @functools.wraps(fn)
        def inner_wrapper(input: torch.Tensor, *args: Any, **kwargs: Any) -> Any:
            if preprocessor is not None:
                input = preprocessor(input)

            output = fn(input, *args, **kwargs)

            if postprocessor is not None:
                output = postprocessor(output)

            return output

        return inner_wrapper

    return outer_wrapper


class TorchPreprocessing(ComposedTransform):
    def __init__(self) -> None:
        transforms = (Normalize(TORCH_MEAN, TORCH_STD),)
        super().__init__(*transforms)


class TorchPostprocessing(ComposedTransform):
    def __init__(self) -> None:
        transforms = (Denormalize(TORCH_MEAN, TORCH_STD),)
        super().__init__(*transforms)


def torch_processing(fn: Callable) -> Callable:
    preprocessor = TorchPreprocessing()
    postprocessor = TorchPostprocessing()

    processing = _processing(preprocessor, postprocessor)
    return processing(fn)


class CaffePreprocessing(ComposedTransform):
    def __init__(self) -> None:
        transforms = (
            Normalize(CAFFE_MEAN, CAFFE_STD),
            FloatToUint8Range(),
            ReverseChannelOrder(),
        )
        super().__init__(*transforms)


class CaffePostprocessing(ComposedTransform):
    def __init__(self) -> None:
        transforms = (
            ReverseChannelOrder(),
            Uint8ToFloatRange(),
            Denormalize(CAFFE_MEAN, CAFFE_STD),
        )
        super().__init__(*transforms)


def caffe_processing(fn: Callable) -> Callable:
    preprocessor = CaffePreprocessing()
    postprocessor = CaffePostprocessing()

    processing = _processing(preprocessor, postprocessor)
    return processing(fn)



import contextlib
import logging
import sys
import time
import warnings
from datetime import datetime
from typing import Callable, Iterator, Optional, Tuple, Union, cast

import torch
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler
from torch.optim.optimizer import Optimizer

import pystiche
from pystiche.misc import build_deprecation_message
from pystiche.pyramid.level import PyramidLevel

from .meter import AverageMeter, ETAMeter, LossMeter, ProgressMeter


def _deprecation_warning() -> None:
    msg = build_deprecation_message(
        "Using any functionality from pystiche.optim.log",
        "0.7.0",
        info="See https://github.com/pmeier/pystiche/issues/434 for details.",
    )
    warnings.warn(msg, UserWarning)


__all__ = [
    "default_logger",
    "OptimLogger",
    "default_image_optim_log_fn",
    "default_pyramid_level_header",
    "default_transformer_optim_log_fn",
    "default_epoch_header",
]


def default_logger(
    name: Optional[str] = None, log_file: Optional[str] = None
) -> logging.Logger:
    r"""Logs :attr:`logging.INFO` to :class:`sys.stdout` and all log levels above to
    :class:`sys.stderr`. Each log entry is timestamped and can optionally be written to
    a log_file.

    Args:
        name: Optional name. See :class:`logging.Logger` for details.
        log_file: Optional path to log file.
    """
    _deprecation_warning()
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    fmt = logging.Formatter(
        fmt="|%(asctime)s| %(message)s", datefmt="%d.%m.%Y %H:%M:%S"
    )

    sh = logging.StreamHandler(sys.stdout)
    sh.setLevel(logging.INFO)
    sh.addFilter(lambda record: record.levelno <= logging.INFO)
    sh.setFormatter(fmt)
    logger.addHandler(sh)

    sh = logging.StreamHandler(sys.stderr)
    sh.setLevel(logging.WARNING)
    sh.setFormatter(fmt)
    logger.addHandler(sh)

    if log_file is not None:
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.INFO)
        fh.setFormatter(fmt)
        logger.addHandler(fh)

    return logger


class OptimLogger:
    r"""Hierarchical optimization logger.

    Args:
        logger: Optional :class:`logging.Logger` that is used. Defaults to
            :func:`default_logger` with a unique name.
    """
    INDENT = 2
    SEP_LINE_LENGTH = 80
    SEP_CHARS = ("#", "*", "=", "-", "^", '"')

    def __init__(self, logger: Optional[logging.Logger] = None):
        _deprecation_warning()
        if logger is None:
            name = f"pystiche_{datetime.now().strftime('%Y%m%d%H%M%S%f')}"
            logger = default_logger(name=name)
        self.logger = logger

        self._environ_indent_offset = 0
        self._environ_level_offset = 0

    def _calc_abs_indent(self, indent: int, rel: bool) -> int:
        abs_indent = indent
        if rel:
            abs_indent += self._environ_indent_offset
        return abs_indent

    def _calc_abs_level(self, level: int, rel: bool) -> int:
        abs_level = level
        if rel:
            abs_level += self._environ_level_offset
        return abs_level

    def message(self, msg: str, indent: int = 0, rel: bool = True) -> None:
        r"""Log a message with optional indent.

        Args:
            msg: Message.
            indent: Optional indent. Defaults to ``0``.
            rel: If ``True``, the indentation is determined relative to the current
                level. Defaults to ``True``.
        """
        abs_indent = self._calc_abs_indent(indent, rel)
        for line in msg.splitlines():
            self.logger.info(" " * abs_indent + line)

    def sepline(self, level: int = 0, rel: bool = True) -> None:
        r"""Log a separation line.

        Args:
            level: Determines the separation char following
                `Python's Style Guide for documenting <https://devguide.python.org/documenting/#sections>`_
                .
            rel: If ``True``, the level is determined relative to the current level.
                Defaults to ``True``.
        """
        abs_level = self._calc_abs_level(level, rel)
        self.message(self.SEP_CHARS[abs_level] * self.SEP_LINE_LENGTH)

    def sep_message(
        self,
        msg: str,
        level: int = 0,
        rel: bool = True,
        top_sep: bool = True,
        bottom_sep: bool = True,
    ) -> None:
        r"""Log a message with optional enclosing separation lines.

        Args:
            msg: Message.
            level: Determines the separation char following
                `Python's Style Guide for documenting <https://devguide.python.org/documenting/#sections>`_
                .
            rel: If ``True``, the level is determined relative to the current level.
                Defaults to ``True``.
            top_sep: If ``True``, add a separation line before the message. Defaults to
                ``True``.
            bottom_sep: If ``True``, add a separation line after the message. Defaults
                to ``True``.
        """
        if top_sep:
            self.sepline(level=level, rel=rel)
        self.message(msg, rel=rel)
        if bottom_sep:
            self.sepline(level=level, rel=rel)

    @contextlib.contextmanager
    def environment(self, header: Optional[str]) -> Iterator:
        r"""Context manager that increases the current level. Can be nested.

        Args:
            header: Optional header that is logged with :meth:`.sep_message`.
        """
        if header is not None:
            self.sep_message(header)
        self._environ_indent_offset += self.INDENT
        self._environ_level_offset += 1
        try:
            yield
        finally:
            self._environ_level_offset -= 1
            self._environ_indent_offset -= self.INDENT


def default_image_optim_log_fn(
    optim_logger: OptimLogger, log_freq: int = 50, max_depth: int = 1
) -> Callable[[int, Union[torch.Tensor, pystiche.LossDict]], None]:
    r"""Log the loss during image optimizations.

    Args:
        optim_logger: Optimization logger.
        log_freq: Number of optimization steps between two log entries. Defaults to
            ``50``.
        max_depth: If loss is :class:`pystiche.LossDict`, aggregate it to this maximum
            depth before logging. Defaults to ``1``.
    """
    _deprecation_warning()

    def log_fn(step: int, loss: Union[torch.Tensor, pystiche.LossDict]) -> None:
        if step % log_freq == 0:
            with optim_logger.environment(f"Step {step}"):
                if isinstance(loss, pystiche.LossDict):
                    loss = loss.aggregate(max_depth)

                if isinstance(loss, pystiche.LossDict):
                    msg = str(loss)
                elif isinstance(loss, torch.Tensor):
                    msg = f"loss: {loss.item():.3e}"
                else:
                    msg = (  # type: ignore[unreachable]
                        f"loss can be a pystiche.LossDict or a scalar torch.Tensor, "
                        f"but got {type(loss)}."
                    )
                    raise TypeError(msg)

                optim_logger.message(msg)

    return log_fn


def default_pyramid_level_header(
    num: int, level: PyramidLevel, input_image_size: Tuple[int, int]
) -> str:
    r"""
    Args:
        num: Number of the pyramid level.
        level: Pyramid level.
        input_image_size: Image size of the input image.

    Returns:
        Header that includes information about the current level.
    """
    _deprecation_warning()
    height, width = input_image_size
    return f"Pyramid level {num} with {level.num_steps} steps " f"({width} x {height})"


def default_transformer_optim_log_fn(
    optim_logger: OptimLogger,
    num_batches: int,
    log_freq: Optional[int] = None,
    show_loading_velocity: bool = True,
    show_processing_velocity: bool = True,
    show_running_means: bool = True,
) -> Callable[[int, Union[float, torch.Tensor, pystiche.LossDict], float, float], None]:
    r"""Log the ETA, the loss, and the loading and processing velocities during
    transformer optimizations.

    Args:
        optim_logger: Optimization logger.
        num_batches: Total number of batches.
        log_freq: Number of batches between two log entries. If ``None`` a sensible
            value is chosen based on ``num_batches``. Defaults to ``None``.
        show_loading_velocity: If ``True``, includes the image loading velocity in the
            log entries. Defaults to ``True``.
        show_processing_velocity: If ``True``, includes the image processing velocity
            in the log entries. Defaults to ``True``.
        show_running_means: If ``True``, includes a running mean additional to the
            current values in the log entries. The window size is chosen based on
            ``log_freq``. Defaults to ``True``.
    """
    _deprecation_warning()
    if log_freq is None:
        log_freq = max(min(round(1e-3 * num_batches) * 10, 50), 1)

    window_size = min(10 * log_freq, 1000)

    meters = [
        ETAMeter(
            num_batches, window_size=window_size, show_local_eta=show_running_means
        ),
        LossMeter(window_size=window_size, show_local_avg=show_running_means),
    ]
    if show_loading_velocity:
        meters.append(
            AverageMeter(
                name="loading",
                window_size=window_size,
                show_local_avg=show_running_means,
                fmt="{:4.0f} img/s",
            )
        )
    if show_processing_velocity:
        meters.append(
            AverageMeter(
                name="processing",
                window_size=window_size,
                show_local_avg=show_running_means,
                fmt="{:4.0f} img/s",
            )
        )

    progress_meter = ProgressMeter(num_batches, *meters)

    def log_fn(
        batch: int,
        loss: Union[float, torch.Tensor, pystiche.LossDict],
        loading_velocity: float,
        processing_velocity: float,
    ) -> None:
        progress_meter.update(
            ETA=time.time(),
            loss=loss,
            loading=loading_velocity,
            processing=processing_velocity,
        )
        if batch % cast(int, log_freq) == 0:
            optim_logger.message(str(progress_meter))

    return log_fn


def default_epoch_header(
    epoch: int, optimizer: Optimizer, lr_scheduler: Optional[LRScheduler]
) -> str:
    r"""
    Args:
        epoch: Number of the epoch.
        optimizer: Optimizer.
        lr_scheduler: Optional learning rate scheduler.

    Returns:
        Header that includes information about the current epoch.
    """
    _deprecation_warning()
    return f"Epoch {epoch}"



from typing import Any, Dict, Sequence, Union

import torch
from torch import nn

__all__ = [
    "tensor_meta",
    "is_scalar_tensor",
    "ConvModule",
    "is_conv_module",
    "conv_module_meta",
    "pool_module_meta",
    "PoolModule",
    "is_pool_module",
]


def _extract_meta_attrs(
    obj: Any, attrs: Sequence[str], **kwargs: Any
) -> Dict[str, Any]:
    for attr in attrs:
        if attr not in kwargs:
            kwargs[attr] = getattr(obj, attr)
    return kwargs


def tensor_meta(x: torch.Tensor, **kwargs: Any) -> Dict[str, Any]:
    attrs = ("dtype", "device")
    return _extract_meta_attrs(x, attrs, **kwargs)


def is_scalar_tensor(x: torch.Tensor) -> bool:
    return x.dim() == 0


ConvModule = Union[
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose2d,
]


def is_conv_module(x: Any) -> bool:
    return isinstance(
        x,
        (
            nn.Conv1d,
            nn.Conv2d,
            nn.Conv3d,
            nn.ConvTranspose1d,
            nn.ConvTranspose2d,
            nn.ConvTranspose3d,
        ),
    )


def conv_module_meta(x: ConvModule, **kwargs: Any) -> Dict[str, Any]:
    attrs = ("kernel_size", "stride", "padding", "dilation")
    return _extract_meta_attrs(x, attrs, **kwargs)


PoolModule = Union[
    nn.AdaptiveAvgPool1d,
    nn.AdaptiveAvgPool2d,
    nn.AdaptiveAvgPool3d,
    nn.AdaptiveMaxPool1d,
    nn.AdaptiveMaxPool2d,
    nn.AdaptiveMaxPool3d,
    nn.AvgPool1d,
    nn.AvgPool2d,
    nn.AvgPool3d,
    nn.MaxPool1d,
    nn.MaxPool2d,
    nn.MaxPool3d,
]


def is_pool_module(x: Any) -> bool:
    return isinstance(
        x,
        (
            nn.AdaptiveAvgPool1d,
            nn.AdaptiveAvgPool2d,
            nn.AdaptiveAvgPool3d,
            nn.AdaptiveMaxPool1d,
            nn.AdaptiveMaxPool2d,
            nn.AdaptiveMaxPool3d,
            nn.AvgPool1d,
            nn.AvgPool2d,
            nn.AvgPool3d,
            nn.MaxPool1d,
            nn.MaxPool2d,
            nn.MaxPool3d,
        ),
    )


def pool_module_meta(x: PoolModule, **kwargs: Any) -> Dict[str, Any]:
    attrs = ("kernel_size", "stride", "padding")
    return _extract_meta_attrs(x, attrs, **kwargs)



from typing import NoReturn, Optional, Tuple, Union, cast

import numpy as np

import torch

from pystiche.image.utils import (
    extract_batch_size,
    extract_image_size,
    extract_num_channels,
    force_batched_image,
)
from pystiche.misc import to_2d_arg, verify_str_arg

from ._align_corners import affine_grid, grid_sample

__all__ = [
    "transform_motif_affinely",
    "shear_motif",
    "rotate_motif",
    "scale_motif",
    "translate_motif",
]


def _create_motif_shearing_matrix(
    angle: float, clockwise: bool = False
) -> torch.Tensor:
    angle = np.deg2rad(angle)
    if clockwise:
        angle *= -1.0
    shearing_matrix = (
        (1.0, -np.sin(angle), 0.0),
        (0.0, np.cos(angle), 0.0),
        (0.0, 0.0, 1.0),
    )
    return torch.tensor(shearing_matrix, dtype=torch.float32)


def _create_motif_rotation_matrix(
    angle: float, clockwise: bool = False
) -> torch.Tensor:
    angle = np.deg2rad(angle)
    if clockwise:
        angle *= -1.0
    rotation_matrix = (
        (np.cos(angle), -np.sin(angle), 0.0),
        (np.sin(angle), np.cos(angle), 0.0),
        (0.0, 0.0, 1.0),
    )
    return torch.tensor(rotation_matrix, dtype=torch.float32)


def _create_motif_scaling_matrix(
    factor: Union[float, Tuple[float, float]]
) -> torch.Tensor:
    factor_vert, factor_horz = to_2d_arg(factor)
    scaling_matrix = ((factor_horz, 0.0, 0.0), (0.0, factor_vert, 0.0), (0.0, 0.0, 1.0))
    return torch.tensor(scaling_matrix, dtype=torch.float32)


def _create_motif_translation_matrix(
    translation: Tuple[float, float], inverse: bool = False
) -> torch.Tensor:
    translation_vert, translation_horz = translation
    if inverse:
        translation_vert = -translation_vert
        translation_horz = -translation_horz
    translation_matrix = (
        (1.0, 0.0, translation_horz),
        (0.0, 1.0, translation_vert),
        (0.0, 0.0, 1.0),
    )
    return torch.tensor(translation_matrix, dtype=torch.float32)


def _calculate_image_center(image_size: Tuple[int, int]) -> Tuple[float, float]:
    height, width = image_size
    vert_center = height / 2.0
    horz_center = width / 2.0
    return vert_center, horz_center


def _transform_around_point(
    point: Tuple[float, float], transform_matrix: torch.Tensor
) -> torch.Tensor:
    return cast(
        torch.Tensor,
        torch.chain_matmul(
            _create_motif_translation_matrix(point, inverse=False),
            transform_matrix,
            _create_motif_translation_matrix(point, inverse=True),
        ),
    )


def _transform_coordinates(
    transform_matrix: torch.Tensor, image_size: Tuple[int, int]
) -> torch.Tensor:
    vert_center, horz_center = _calculate_image_center(image_size)
    coordinate_transform_matrix = (
        (horz_center, 0.0, horz_center),
        (0.0, -vert_center, vert_center),
        (0.0, 0.0, 1.0),
    )
    coordinate_transform_matrix = torch.tensor(coordinate_transform_matrix)
    return cast(
        torch.Tensor,
        torch.chain_matmul(
            torch.inverse(coordinate_transform_matrix),
            transform_matrix,
            coordinate_transform_matrix,
        ),
    )


def _create_affine_transform_matrix(
    image_size: Tuple[int, int],
    shearing_angle: Optional[float] = None,
    clockwise_shearing: bool = False,
    shearing_center: Optional[Tuple[float, float]] = None,
    rotation_angle: Optional[float] = None,
    clockwise_rotation: bool = False,
    rotation_center: Optional[Tuple[float, float]] = None,
    scaling_factor: Optional[Union[float, Tuple[float, float]]] = None,
    scaling_center: Optional[Tuple[float, float]] = None,
    translation: Optional[Tuple[float, float]] = None,
    inverse_translation: bool = False,
) -> torch.Tensor:
    if not any((shearing_angle, rotation_angle, scaling_factor, translation)):
        return torch.eye(3)

    image_center = _calculate_image_center(image_size)
    if shearing_center is None:
        shearing_center = image_center
    if rotation_center is None:
        rotation_center = image_center
    if scaling_center is None:
        scaling_center = image_center

    transform_matrices = []
    if shearing_angle is not None:
        transform_matrix = _create_motif_shearing_matrix(
            shearing_angle, clockwise=clockwise_shearing
        )
        transform_matrix = _transform_around_point(shearing_center, transform_matrix)
        transform_matrices.append(transform_matrix)
    if rotation_angle is not None:
        transform_matrix = _create_motif_rotation_matrix(
            rotation_angle, clockwise=clockwise_rotation
        )
        transform_matrix = _transform_around_point(rotation_center, transform_matrix)
        transform_matrices.append(transform_matrix)
    if scaling_factor is not None:
        transform_matrix = _create_motif_scaling_matrix(scaling_factor)
        transform_matrix = _transform_around_point(scaling_center, transform_matrix)
        transform_matrices.append(transform_matrix)
    if translation is not None:
        transform_matrix = _create_motif_translation_matrix(
            translation, inverse=inverse_translation
        )
        transform_matrices.append(transform_matrix)

    return cast(torch.Tensor, torch.chain_matmul(*reversed(transform_matrices)))


def _calculate_full_bounding_box_size(vertices: torch.Tensor) -> Tuple[int, int]:
    # When calling torch.max with a dimension as second argument, a namedtuple with a
    # values field is returned.
    # https://pytorch.org/docs/stable/torch.html#torch.max
    # This is not reflected in the type hints.
    coords = torch.max(torch.abs(vertices), 1).values
    return cast(
        Tuple[int, int],
        tuple(reversed(torch.ceil(2.0 * coords).to(torch.int).tolist())),
    )


def _calculate_valid_bounding_box_size(vertices: torch.Tensor) -> NoReturn:
    msg = "The valid canvas option is not yet implemented."
    raise RuntimeError(msg)


def _resize_canvas(
    transform_matrix: torch.Tensor, image_size: Tuple[int, int], method: str = "same",
) -> Tuple[torch.Tensor, Tuple[int, int]]:
    verify_str_arg(method, "method", ("same", "full", "valid"))

    if method == "same":
        return transform_matrix, image_size

    def center_motif(
        transform_matrix: torch.Tensor, image_size: Tuple[int, int]
    ) -> torch.Tensor:
        image_center = _calculate_image_center(image_size)
        image_center = torch.tensor((*image_center[::-1], 1.0)).unsqueeze(1)
        motif_center = torch.mm(transform_matrix, image_center)
        motif_center = cast(Tuple[float, float], motif_center[:-1, 0].tolist()[::-1])

        translation_matrix = _create_motif_translation_matrix(
            motif_center, inverse=True
        )
        return torch.mm(translation_matrix, transform_matrix)

    def calculate_motif_vertices(
        transform_matrix: torch.Tensor, image_size: Tuple[int, int]
    ) -> torch.Tensor:
        height, width = image_size
        # TODO: do this without transpose
        image_vertices = torch.tensor(
            (
                (0.0, 0.0, 1.0),
                (width, 0.0, 1.0),
                (0.0, height, 1.0),
                (width, height, 1.0),
            )
        ).t()
        return torch.mm(transform_matrix, image_vertices)[:-1, :]

    def scale_and_off_center_motif(
        transform_matrix: torch.Tensor,
        image_size: Tuple[int, int],
        bounding_box_size: Tuple[int, int],
    ) -> torch.Tensor:
        height, width = image_size
        image_center = _calculate_image_center(image_size)

        bounding_box_height, bounding_box_width = bounding_box_size
        scaling_factors = (height / bounding_box_height, width / bounding_box_width)
        scaling_matrix = _create_motif_scaling_matrix(scaling_factors)
        scaling_matrix = _transform_around_point(image_center, scaling_matrix)

        translation_matrix = _create_motif_translation_matrix(image_center)

        return cast(
            torch.Tensor,
            torch.chain_matmul(scaling_matrix, translation_matrix, transform_matrix),
        )

    transform_matrix = center_motif(transform_matrix, image_size)
    motif_vertices = calculate_motif_vertices(transform_matrix, image_size)

    if method == "full":
        bounding_box_size = _calculate_full_bounding_box_size(motif_vertices)
    else:  # method == "valid"
        bounding_box_size = _calculate_valid_bounding_box_size(motif_vertices)

    transform_matrix = scale_and_off_center_motif(
        transform_matrix, image_size, bounding_box_size
    )

    return transform_matrix, bounding_box_size


def _calculate_affine_grid(
    image: torch.Tensor, transform_matrix: torch.Tensor, image_size: Tuple[int, int]
) -> torch.Tensor:
    batch_size = extract_batch_size(image)
    num_channels = extract_num_channels(image)
    height, width = image_size
    size = (batch_size, num_channels, height, width)

    inv_transform_matrix = torch.inverse(transform_matrix)
    theta = inv_transform_matrix[:-1, :].unsqueeze(0)

    return affine_grid(theta, size)


@force_batched_image
def transform_motif_affinely(
    image: torch.Tensor,
    shearing_angle: Optional[float] = None,
    clockwise_shearing: bool = False,
    shearing_center: Optional[Tuple[float, float]] = None,
    rotation_angle: Optional[float] = None,
    clockwise_rotation: bool = False,
    rotation_center: Optional[Tuple[float, float]] = None,
    scaling_factor: Optional[Union[float, Tuple[float, float]]] = None,
    scaling_center: Optional[Tuple[float, float]] = None,
    translation: Optional[Tuple[float, float]] = None,
    inverse_translation: bool = False,
    canvas: str = "same",
    interpolation_mode: str = "bilinear",
    padding_mode: str = "zeros",
) -> torch.Tensor:
    device = image.device
    image_size = extract_image_size(image)

    transform_matrix = _create_affine_transform_matrix(
        image_size,
        shearing_angle=shearing_angle,
        clockwise_shearing=clockwise_shearing,
        shearing_center=shearing_center,
        rotation_angle=rotation_angle,
        clockwise_rotation=clockwise_rotation,
        rotation_center=rotation_center,
        scaling_factor=scaling_factor,
        scaling_center=scaling_center,
        translation=translation,
        inverse_translation=inverse_translation,
    )
    transform_matrix, resized_image_size = _resize_canvas(
        transform_matrix, image_size, method=canvas
    )
    transform_matrix = _transform_coordinates(transform_matrix, image_size)
    transform_matrix = transform_matrix.to(device)

    grid = _calculate_affine_grid(image, transform_matrix, resized_image_size)
    grid = grid.to(device)

    return grid_sample(image, grid, mode=interpolation_mode, padding_mode=padding_mode,)


def shear_motif(
    image: torch.Tensor,
    angle: float,
    clockwise: bool = False,
    center: Optional[Tuple[float, float]] = None,
    canvas: str = "same",
    interpolation_mode: str = "bilinear",
    padding_mode: str = "zeros",
) -> torch.Tensor:
    return cast(
        torch.Tensor,
        transform_motif_affinely(
            image,
            shearing_angle=angle,
            clockwise_shearing=clockwise,
            shearing_center=center,
            canvas=canvas,
            interpolation_mode=interpolation_mode,
            padding_mode=padding_mode,
        ),
    )


def rotate_motif(
    image: torch.Tensor,
    angle: float,
    clockwise: bool = False,
    center: Optional[Tuple[float, float]] = None,
    canvas: str = "same",
    interpolation_mode: str = "bilinear",
    padding_mode: str = "zeros",
) -> torch.Tensor:
    return cast(
        torch.Tensor,
        transform_motif_affinely(
            image,
            rotation_angle=angle,
            clockwise_rotation=clockwise,
            rotation_center=center,
            canvas=canvas,
            interpolation_mode=interpolation_mode,
            padding_mode=padding_mode,
        ),
    )


def scale_motif(
    image: torch.Tensor,
    factor: Union[float, Tuple[float, float]],
    center: Optional[Tuple[float, float]] = None,
    canvas: str = "same",
    interpolation_mode: str = "bilinear",
    padding_mode: str = "zeros",
) -> torch.Tensor:
    return cast(
        torch.Tensor,
        transform_motif_affinely(
            image,
            scaling_factor=factor,
            scaling_center=center,
            canvas=canvas,
            interpolation_mode=interpolation_mode,
            padding_mode=padding_mode,
        ),
    )


def translate_motif(
    image: torch.Tensor,
    translation: Tuple[float, float],
    inverse: bool = False,
    canvas: str = "same",
    interpolation_mode: str = "bilinear",
    padding_mode: str = "zeros",
) -> torch.Tensor:
    return cast(
        torch.Tensor,
        transform_motif_affinely(
            image,
            translation=translation,
            inverse_translation=inverse,
            canvas=canvas,
            interpolation_mode=interpolation_mode,
            padding_mode=padding_mode,
        ),
    )



import os
from os import path
from typing import Any, Dict, Optional, Union

import torch
from torch import nn
from torchvision.datasets.utils import check_md5

import pystiche
from pystiche.misc import download_file

from ..license import License, UnknownLicense
from ._core import _Image, _ImageCollection


class DownloadableImage(_Image):
    def __init__(
        self,
        url: str,
        title: Optional[str] = None,
        author: Optional[str] = None,
        date: Optional[str] = None,
        license: Optional[Union[License, str]] = None,
        md5: Optional[str] = None,
        file: Optional[str] = None,
        guides: Optional["DownloadableImageCollection"] = None,
        prefix_guide_files: bool = True,
        transform: Optional[nn.Module] = None,
        note: Optional[str] = None,
    ) -> None:
        r"""Downloadable image.

        Args:
            url: URL to the image.
            title: Optional title of the image.
            author: Optional author of the image.
            date: Optional date of the image.
            license: Optional license of the image.
            md5: Optional `MD5 <https://en.wikipedia.org/wiki/MD5>`_ checksum of the
                image file.
            file: Optional path to the image file. If ``None``, see
                :meth:`.generate_file` for details.
            guides: Optional guides for the image.
            prefix_guide_files: If ``True``, the guide files are prefixed with the
                ``file`` name.
            transform: Optional transform that is applied to the image after it is
                :meth:`~.read`.
            note: Optional note that is included in the representation.
        """
        if file is None:
            file = self.generate_file(url, title, author)

        if guides is not None and prefix_guide_files:
            prefix = path.splitext(path.basename(file))[0]
            for _, guide in guides:
                if not path.isabs(guide.file):
                    guide.file = path.join(prefix, guide.file)

        super().__init__(file, guides=guides, transform=transform, note=note)
        self.url = url
        self.title = title
        self.author = author
        self.date = date

        if license is None:
            license = UnknownLicense()
        self.license = license

        self.md5 = md5

    @staticmethod
    def generate_file(url: str, title: Optional[str], author: Optional[str]) -> str:
        r"""Generate a filename from the supplied information from the following scheme:

        - If ``title`` and ``author`` are ``None``, the ending of ``url`` is used.
        - If one of ``title`` or ``author`` is not ``None``, it is used as filename where spaces are replaced by underscores.
        - If ``title`` and ``author`` are not None, the filename is generated as above separating both path with double underscores.

        Args:
            url: URL to the image.
            title: Optional title of the image.
            author: Optional author of the image
        """
        if title is None and author is None:
            return path.basename(url)

        def format(x: str) -> str:
            return "_".join(x.lower().split())

        name_parts = [format(part) for part in (title, author) if part is not None]
        name = "__".join(name_parts)
        ext = path.splitext(url)[1]
        return name + ext

    def download(self, root: Optional[str] = None, overwrite: bool = False) -> None:
        r"""Download the image and if applicable the guides from their URL. If the
        correct MD5 checksum is known, it is verified first. If it checks out the file
        not re-downloaded.

        Args:
            root: Optional root directory for the download if the file is a relative
                path. Defaults to :func:`pystiche.home`.
            overwrite: Overwrites files if they already exists or the MD5 checksum does
                not match. Defaults to ``False``.
        """

        def _download(file: str) -> None:
            os.makedirs(path.dirname(file), exist_ok=True)
            download_file(self.url, file=file, md5=self.md5)

        if root is None:
            root = pystiche.home()

        if isinstance(self.guides, DownloadableImageCollection):
            self.guides.download(root=root, overwrite=overwrite)

        file = self.file
        if not path.isabs(file) and root is not None:
            file = path.join(root, file)

        if not path.isfile(file):
            _download(file)
            return

        msg_overwrite = "If you want to overwrite it, set overwrite=True."

        if self.md5 is None:
            if overwrite:
                _download(file)
                return
            else:
                msg = f"{path.basename(file)} already exists in {root}. {msg_overwrite}"
                raise FileExistsError(msg)

        if not check_md5(file, self.md5):
            if overwrite:
                _download(file)
                return
            else:
                msg = (
                    f"{path.basename(file)} with a different MD5 hash already exists "
                    f"in {root}. {msg_overwrite}"
                )
                raise FileExistsError(msg)

    def read(
        self,
        root: Optional[str] = None,
        download: Optional[bool] = None,
        overwrite: bool = False,
        **read_image_kwargs: Any,
    ) -> torch.Tensor:
        r"""Read the image from file with :func:`pystiche.image.read_image`. If
        available the :attr:`.transform` is applied afterwards.

        Args:
            root: Optional root directory if the file is a relative path.
                Defaults to :func:`pystiche.home`.
            download: If ``True``, downloads the image first. Defaults to ``False`` if
                the file already exists and the MD5 checksum is not known. Otherwise
                defaults to ``True``.
            overwrite: If downloaded, overwrites files if they already exists or the
                MD5 checksum does not match. Defaults to ``False``.
            **read_image_kwargs: Optional parameters passed to
                :func:`pystiche.image.read_image`.
        """
        if root is None:
            root = pystiche.home()
        if download is None:
            file_exists = path.isfile(path.join(root, self.file))
            md5_available = self.md5 is not None
            download = False if file_exists and not md5_available else True
        if download:
            self.download(root=root, overwrite=overwrite)
        return super().read(root=root, **read_image_kwargs)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["url"] = self.url
        dct["title"] = self.title if self.title is not None else "unknown"
        dct["author"] = self.author if self.author is not None else "unknown"
        dct["date"] = self.date if self.date is not None else "unknown"
        dct["license"] = self.license
        return dct


class DownloadableImageCollection(_ImageCollection):
    def download(self, root: Optional[str] = None, overwrite: bool = False) -> None:
        r"""Download all images and if applicable their guides from their URLs. See
        :meth:`pystiche.data.DownloadableImage.download` for details.

        Args:
            root: Optional root directory for the download if the file is a relative
                path. Defaults to :func:`pystiche.home`.
            overwrite: Overwrites files if they already exists or the MD5 checksum does
                not match. Defaults to ``False``.
        """
        for _, image in self:
            if isinstance(image, DownloadableImage):
                image.download(root=root, overwrite=overwrite)

    def read(
        self,
        root: Optional[str] = None,
        download: Optional[bool] = None,
        overwrite: bool = False,
        **read_image_kwargs: Any,
    ) -> Dict[str, torch.Tensor]:
        r"""Read the images from file. See :meth:`pystiche.data.DownloadableImage.read`
        for details.

        Args:
            root: Optional root directory if the file is a relative path.
                Defaults to :func:`pystiche.home`.
            download: If ``True``, downloads the image first. Defaults to ``False`` if
                the file already exists and the MD5 checksum is not known. Otherwise
                defaults to ``True``.
            overwrite: If downloaded, overwrites files if they already exists or the
                MD5 checksum does not match. Defaults to ``False``.
            **read_image_kwargs: Optional parameters passed to
                :func:`pystiche.image.read_image`.

        Returns:
            Dictionary with the name image pairs.
        """
        return {
            name: image.read(
                root=root, download=download, overwrite=overwrite, **read_image_kwargs
            )
            for name, image in self
        }



from typing import (
    Any,
    Collection,
    Dict,
    Iterator,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import numpy as np

from pystiche import ComplexObject
from pystiche.loss import MultiOperatorLoss
from pystiche.misc import zip_equal
from pystiche.ops import ComparisonOperator, Operator, OperatorContainer

from .level import PyramidLevel
from .storage import ImageStorage

__all__ = ["ImagePyramid", "OctaveImagePyramid"]


class ImagePyramid(ComplexObject):
    r"""Image pyramid for a coarse-to-fine optimization on different levels. If
    iterated on yields :class:`~pystiche.pyramid.PyramidLevel` s and handles the
    resizing of all set images and guides of ``resize_targets``.

    Args:
        edge_sizes: Edge sizes for each level.
        num_steps: Number of steps for each level. If sequence of ``int`` its length
            has to match the length of ``edge_sizes``.
        edge: Corresponding edge to the edge size for each level. Can be ``"short"`` or
            ``"long"``. If sequence of ``str`` its length has to match the length of
            ``edge_sizes``. Defaults to ``"short"``.
        interpolation_mode: Interpolation mode used for the resizing of the images.
            Defaults to ``"bilinear"``.

            .. note::
                For the resizing of guides ``"nearest"`` is used regardless of the
                ``interpolation_mode``.
        resize_targets: Targets for resizing of set images and guides during iteration.
    """

    def __init__(
        self,
        edge_sizes: Sequence[int],
        num_steps: Union[Sequence[int], int],
        edge: Union[Sequence[str], str] = "short",
        interpolation_mode: str = "bilinear",
        resize_targets: Collection[Union[Operator, MultiOperatorLoss]] = (),
    ):
        self._levels = self.build_levels(edge_sizes, num_steps, edge)
        self.interpolation_mode = interpolation_mode
        self._resize_targets = set(resize_targets)

    @staticmethod
    def build_levels(
        edge_sizes: Sequence[int],
        num_steps: Union[Sequence[int], int],
        edge: Union[Sequence[str], str],
    ) -> Tuple[PyramidLevel, ...]:
        num_levels = len(edge_sizes)
        if isinstance(num_steps, int):
            num_steps = [num_steps] * num_levels
        if isinstance(edge, str):
            edge = [edge] * num_levels

        return tuple(
            PyramidLevel(edge_size, num_steps_, edge_)
            for edge_size, num_steps_, edge_ in zip_equal(edge_sizes, num_steps, edge)
        )

    # TODO: can this be removed?
    def add_resize_target(self, op: Operator) -> None:
        self._resize_targets.add(op)

    def __len__(self) -> int:
        return len(self._levels)

    def __getitem__(self, idx: int) -> PyramidLevel:
        return self._levels[idx]

    def __iter__(self) -> Iterator[PyramidLevel]:
        image_storage = ImageStorage(self._resize_ops())
        for level in self._levels:
            try:
                self._resize(level)
                yield level
            finally:
                image_storage.restore()

    def _resize(self, level: PyramidLevel) -> None:
        for op in self._resize_ops():
            if isinstance(op, ComparisonOperator):
                if op.has_target_guide:
                    resized_guide = level.resize_guide(op.target_guide)
                    op.set_target_guide(resized_guide, recalc_repr=False)

                if op.has_target_image:
                    resized_image = level.resize_image(
                        op.target_image, interpolation_mode=self.interpolation_mode
                    )
                    op.set_target_image(resized_image)

            if op.has_input_guide:
                resized_guide = level.resize_guide(op.input_guide)
                op.set_input_guide(resized_guide)

    def _resize_ops(self) -> Set[Operator]:
        resize_ops = set()
        for target in self._resize_targets:
            if isinstance(target, Operator):
                resize_ops.add(target)

            for op in target.operators(recurse=True):
                if not isinstance(op, OperatorContainer):
                    resize_ops.add(op)
        return resize_ops

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        return dct

    def _named_children(self) -> Iterator[Tuple[str, Any]]:
        yield from super()._named_children()
        for idx, level in enumerate(self._levels):
            yield str(idx), level


class OctaveImagePyramid(ImagePyramid):
    r"""Image pyramid that comprises levels spaced by a factor of two.

    Args:
        max_edge_size: Maximum edge size.
        num_steps: Number of steps for each level.

            .. note::
                If ``num_steps`` is specified as sequence of ``int``s, you should also
                specify ``num_levels`` to match the lengths
        num_levels: Optional number of levels. If ``None``, the number is determined by
            the number of steps of factor two between ``max_edge_size`` and
            ``min_edge_size``.
        min_edge_size: Minimum edge size for the automatic calculation of
            ``num_levels``.
        image_pyramid_kwargs: Additional options. See
            :class:`~pystiche.pyramid.ImagePyramid` for details.
    """

    def __init__(
        self,
        max_edge_size: int,
        num_steps: Union[int, Sequence[int]],
        num_levels: Optional[int] = None,
        min_edge_size: int = 64,
        **image_pyramid_kwargs: Any,
    ) -> None:
        if num_levels is None:
            num_levels = int(np.floor(np.log2(max_edge_size / min_edge_size))) + 1

        edge_sizes = [
            round(max_edge_size / (2.0 ** ((num_levels - 1) - level)))
            for level in range(num_levels)
        ]
        super().__init__(edge_sizes, num_steps, **image_pyramid_kwargs)



from typing import Optional

import torch
from torch.nn.functional import mse_loss, relu

import pystiche
from pystiche.misc import reduce


def mrf_loss(
    input: torch.Tensor,
    target: torch.Tensor,
    eps: float = 1e-8,
    reduction: str = "mean",
    batched_input: Optional[bool] = None,
) -> torch.Tensor:
    r"""Calculates the MRF loss. See :class:`pystiche.ops.MRFOperator` for details.

    Args:
        input: Input of shape :math:`S_1 \times N_1 \times \dots \times N_D`.
        target: Target of shape :math:`S_2 \times N_1 \times \dots \times N_D`.
        eps: Small value to avoid zero division. Defaults to ``1e-8``.
        reduction: Reduction method of the output passed to
            :func:`pystiche.misc.reduce`. Defaults to ``"mean"``.
        batched_input: If ``True``, treat the first dimension of the inputs as batch
            dimension, i.e. :math:`B \times S \times N_1 \times \dots \times N_D`.
            Defaults to ``False``. See :func:`pystiche.cosine_similarity` for details.

    Note:
        The default value of ``batched_input`` will change from ``False`` to ``True``
        in the future.

    Examples:

        >>> input = torch.rand(256, 64, 3, 3)
        >>> target = torch.rand(256, 64, 3, 3)
        >>> score = F.mrf_loss(input, target)

    """
    with torch.no_grad():
        similarity = pystiche.cosine_similarity(
            input, target, eps=eps, batched_input=batched_input
        )
        idcs = torch.argmax(similarity, dim=1)
        target = torch.index_select(target, dim=0, index=idcs)
    return mse_loss(input, target, reduction=reduction)


def value_range_loss(
    input: torch.Tensor, min: float = 0.0, max: float = 1.0, reduction: str = "mean"
) -> torch.Tensor:
    # TODO: remove sinh call; quadratic, i.e. x * (x-1) is enough
    loss = relu(torch.sinh(input - min) * torch.sinh(input - max))
    return reduce(loss, reduction)


def total_variation_loss(
    input: torch.Tensor, exponent: float = 2.0, reduction: str = "mean"
) -> torch.Tensor:
    r"""Calculates the total variation loss. See
    :class:`pystiche.ops.TotalVariationOperator` for details.

    Args:
        input: Input image
        exponent: Parameter :math:`\beta` . A higher value leads to more smoothed
            results. Defaults to ``2.0``.
        reduction: Reduction method of the output passed to
            :func:`pystiche.misc.reduce`. Defaults to ``"mean"``.

    Examples:

        >>> input = torch.rand(2, 3, 256, 256)
        >>> score = F.total_variation_loss(input)
    """
    # this ignores the last row and column of the image
    grad_vert = input[:, :, :-1, :-1] - input[:, :, 1:, :-1]
    grad_horz = input[:, :, :-1, :-1] - input[:, :, :-1, 1:]
    grad = pystiche.nonnegsqrt(grad_vert ** 2.0 + grad_horz ** 2.0)
    loss = grad ** exponent
    return reduce(loss, reduction)



from typing import Any, Dict, Sequence, cast

import torch

from . import functional as F
from .core import Transform

__all__ = [
    "FloatToUint8Range",
    "Uint8ToFloatRange",
    "ReverseChannelOrder",
    "Normalize",
    "Denormalize",
]


class FloatToUint8Range(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.float_to_uint8_range(x)


class Uint8ToFloatRange(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.uint8_to_float_range(x)


class ReverseChannelOrder(Transform):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, F.reverse_channel_order(x))


def _format_stats(stats: Sequence[float], fmt: str = "{:g}") -> str:
    return str(tuple(fmt.format(stat) for stat in stats))


class Normalize(Transform):
    def __init__(self, mean: Sequence[float], std: Sequence[float]) -> None:
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, F.normalize(x, self.mean, self.std))

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["mean"] = _format_stats(self.mean)
        dct["std"] = _format_stats(self.std)
        return dct


class Denormalize(Transform):
    def __init__(self, mean: Sequence[float], std: Sequence[float]) -> None:
        super().__init__()
        self.mean = mean
        self.std = std

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, F.denormalize(x, self.mean, self.std))

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["mean"] = _format_stats(self.mean)
        dct["std"] = _format_stats(self.std)
        return dct



import itertools
import os
from functools import reduce
from os import path
from typing import Any, Dict, Optional, Tuple, Union, cast

import torch

from .io import read_image, write_image
from .utils import extract_image_size, extract_num_channels, force_single_image

__all__ = [
    "verify_guides",
    "read_guides",
    "write_guides",
    "guides_to_segmentation",
    "segmentation_to_guides",
]

# Color are taken from
# https://colorbrewer2.org/#type=qualitative&scheme=Paired&n=12
# fmt: off
COLOR_ORDER = (
    (166, 206, 227),
    ( 31, 120, 180),  # noqa: E201
    (178, 223, 138),
    ( 51, 160,  44),  # noqa: E201, E241
    (251, 154, 153),
    (227,  26,  28),  # noqa: E241
    (253, 191, 111),
    (255, 127,   0),  # noqa: E241
    (202, 178, 214),
    (106,  61, 154),  # noqa: E241
    (255, 255, 153),
    (177,  89,  40),  # noqa: E241
)
# fmt: on

RGBTriplet = Tuple[int, int, int]


def verify_guides(
    guides: Dict[str, torch.Tensor],
    verify_coverage: bool = True,
    verify_overlap: bool = True,
) -> None:
    """Verify if guides cover the whole canvas and if they do not overlap with each
    other.

    Args:
        guides: Guides to be verified.
        verify_coverage: If ``True`` verifies that the guides cover the whole canvas.
            Defaults to ``True``.
        verify_overlap: If ``True`` verifies that the guides do not overlap with each
            other. Defaults to ``True``.

    Raises:
        RuntimeError: If at least on check is not successful.
    """
    if not (verify_coverage or verify_overlap):
        return

    masks = {region: guide.bool() for region, guide in guides.items()}

    if verify_coverage:
        coverage = reduce(lambda mask1, mask2: mask1 | mask2, masks.values())
        if not torch.all(coverage):
            numel = coverage.numel()
            abs_miss = numel - torch.sum(coverage)
            rel_miss = abs_miss.float() / numel
            msg = f"{abs_miss} pixels ({rel_miss:.1%}) are not covered by the guides."
            raise RuntimeError(msg)

    if verify_overlap:
        overlaps = []
        for (region1, mask1), (region2, mask2) in itertools.combinations(
            masks.items(), 2
        ):
            overlap = mask1 & mask2
            if torch.any(overlap):
                overlaps.append(f"{region1}, {region2}: {torch.sum(overlap)} pixels")

        if overlaps:
            msg = "\n".join(
                (
                    "The guides in the following regions overlap with each other:",
                    "",
                    *overlaps,
                )
            )
            raise RuntimeError(msg)


def read_guides(
    dir: str,
    device: Union[torch.device, str] = "cpu",
    make_batched: bool = True,
    size: Optional[Union[int, Tuple[int, int]]] = None,
    interpolation_mode: str = "nearest",
) -> Dict[str, torch.Tensor]:
    """Read all guides from a directory using :func:`~pystiche.image.read_image` and
    return them as dictionary. The filename without extensions is used as region key.

    Args:
        dir: Path to root directory of the guide files.
        device: Device that the guides are transferred to. Defaults to CPU.
        make_batched: If ``True``, a fake batch dimension is added to every guide.
        size: Optional size the guides are resized to.
        interpolation_mode: Interpolation mode that is used to perform the optional
            resizing. Valid modes are ``"nearest"``, ``"bilinear"``, and ``"bicubic"``.
            Defaults to ``"nearest"``.
    """

    def read_guide(file: str) -> torch.Tensor:
        return read_image(
            path.join(dir, file),
            device=device,
            make_batched=make_batched,
            size=size,
            interpolation_mode=interpolation_mode,
        )

    return {path.splitext(file)[0]: read_guide(file) for file in os.listdir(dir)}


def write_guides(
    guides: Dict[str, torch.Tensor],
    dir: str,
    ext: str = ".png",
    mode: str = "L",
    **save_kwargs: Any,
) -> None:
    """Write guides to directory using :func:`~pystiche.image.write_image`. The region
    key is used as filename.

    Args:
        guides: Guides to be written.
        dir: Path to root directory of the guide files.
        ext: Extension that is appended to the filename. Defaults to ``".png"``.
        mode: Optional image mode. See the `Pillow documentation
            <https://pillow.readthedocs.io/en/stable/handbook/concepts.html#modes>`_
            for details. Defaults to ``"L"``.
        **save_kwargs: Other parameters that are passed to :meth:`PIL.Image.Image.save`
            .
    """
    for region, guide in guides.items():
        file = path.join(dir, region + ext)
        write_image(guide, file, mode=mode, **save_kwargs)


def guides_to_segmentation(
    guides: Dict[str, torch.Tensor], color_map: Optional[Dict[str, RGBTriplet]] = None,
) -> torch.Tensor:
    """Combines multiple guides into one segmentation image.

    Args:
        guides: Guides to be combined.
        color_map: Optional mapping from regions to RGB triplets. If omitted,
            `this color scheme <https://colorbrewer2.org/#type=qualitative&scheme=Paired&n=12>`_
            is used for the region keys in alphabetical order.
    """
    if color_map is None:
        color_map = dict(zip(sorted(guides.keys()), COLOR_ORDER))

    image_size = extract_image_size(tuple(guides.values())[0])
    height, width = image_size
    seg_flat = torch.empty(3, height * width, dtype=torch.float32)

    for region, guide in guides.items():
        rgb_triplet = color_map[region]
        color = torch.tensor(rgb_triplet, dtype=torch.float32).div(255.0).view(3, 1)
        seg_flat[:, guide.bool().flatten()] = color

    return seg_flat.view(1, 3, *image_size)


@force_single_image
def segmentation_to_guides(
    seg: torch.Tensor, region_map: Optional[Dict[RGBTriplet, str]] = None
) -> Dict[Union[RGBTriplet, str], torch.Tensor]:
    """Splits a segmentation image in multiple guides.

    Args:
        seg: Segmentation image to be split.
        region_map: Optional mapping from RGB triplets to regions. If omitted, the RGB
            triplets are used as key in the output.
    """
    num_chanels = extract_num_channels(seg)
    if num_chanels != 3:
        msg = f"Expected 3 channels in the segmentation image, but got {num_chanels}."
        raise ValueError(msg)

    image_size = extract_image_size(seg)

    seg_flat = seg.view(3, -1)
    seg_flat = seg_flat.mul(255.0).byte()

    colors = seg_flat.unique(sorted=False, dim=1).split(1, dim=1)
    guides: Dict[Union[RGBTriplet, str], torch.Tensor] = {}
    for color in colors:
        guide_flat = torch.all(seg_flat == color, dim=0)
        guide = guide_flat.float().view(1, 1, *image_size)

        rgb_triplet = cast(RGBTriplet, tuple(color.squeeze().tolist()))
        guides[rgb_triplet] = guide

    if region_map is None:
        return guides

    return {
        region_map.get(rgb_triplet, rgb_triplet): guide
        for rgb_triplet, guide in cast(Dict[RGBTriplet, torch.Tensor], guides).items()
    }



import functools
from typing import Any, Callable, Sequence, Tuple, cast

import torch

from pystiche.misc import verify_str_arg

__all__ = [
    "verify_is_single_image",
    "is_single_image",
    "verify_is_batched_image",
    "is_batched_image",
    "verify_is_image",
    "is_image",
    "is_image_size",
    "is_edge_size",
    "calculate_aspect_ratio",
    "image_to_edge_size",
    "edge_to_image_size",
    "extract_batch_size",
    "extract_num_channels",
    "extract_image_size",
    "extract_edge_size",
    "extract_aspect_ratio",
    "make_batched_image",
    "make_single_image",
    "force_image",
    "force_single_image",
    "force_batched_image",
]


def _verify_image_type(x: Any) -> None:
    msg = (
        "pystiche uses a torch.Tensor with dtype==torch.float32 "
        "as native image data type, but got input "
    )
    if not isinstance(x, torch.Tensor):
        msg += f"of type {type(x)} instead."
        raise TypeError(msg)
    elif x.dtype != torch.float32:
        msg += f"with dtype=={x.dtype} instead."
        raise TypeError(msg)


def _verify_single_image_dims(x: torch.Tensor) -> None:
    if x.dim() != 3:
        msg = (
            f"pystiche uses CxHxW tensors for single images, but got tensor with "
            f"{x.dim()} dimensions instead."
        )
        raise TypeError(msg)


def _verify_batched_image_dims(x: torch.Tensor) -> None:
    if x.dim() != 4:
        msg = (
            f"pystiche uses BxCxHxW tensors for batched images, but got tensor with "
            f"{x.dim()} dimensions instead."
        )
        raise TypeError(msg)


def _verify_image_dims(x: torch.Tensor) -> None:
    if x.dim() not in (3, 4):
        msg = (
            f"pystiche uses CxHxW tensors for single and BxCxHxW tensors for batched "
            f"images, but got tensor with {x.dim()} dimensions instead."
        )
        raise TypeError(msg)


def verify_is_single_image(x: Any) -> None:
    _verify_image_type(x)
    _verify_single_image_dims(x)


def is_single_image(x: Any) -> bool:
    try:
        verify_is_single_image(x)
    except TypeError:
        return False
    else:
        return True


def verify_is_batched_image(x: Any) -> None:
    _verify_image_type(x)
    _verify_batched_image_dims(x)


def is_batched_image(x: Any) -> bool:
    try:
        verify_is_batched_image(x)
    except TypeError:
        return False
    else:
        return True


def verify_is_image(x: Any) -> None:
    _verify_image_type(x)
    _verify_image_dims(x)


def is_image(x: Any) -> bool:
    try:
        verify_is_image(x)
    except TypeError:
        return False
    else:
        return True


def is_image_size(x: Any) -> bool:
    return (
        isinstance(x, Sequence)
        and len(x) == 2
        and all(map(lambda item: isinstance(item, int), x))
    )


def is_edge_size(x: Any) -> bool:
    return isinstance(x, int)


def calculate_aspect_ratio(image_size: Tuple[int, int]) -> float:
    height, width = image_size
    return width / height


def image_to_edge_size(image_size: Tuple[int, int], edge: str = "short") -> int:
    edge = verify_str_arg(edge, "edge", ("short", "long", "vert", "horz"))
    if edge == "short":
        return min(image_size)
    elif edge == "long":
        return max(image_size)
    elif edge == "vert":
        return image_size[0]
    else:  # edge == "horz"
        return image_size[1]


def edge_to_image_size(
    edge_size: int, aspect_ratio: float, edge: str = "short"
) -> Tuple[int, int]:
    edge = verify_str_arg(edge, "edge", ("short", "long", "vert", "horz"))
    if edge == "vert":
        return edge_size, round(edge_size * aspect_ratio)
    elif edge == "horz":
        return round(edge_size / aspect_ratio), edge_size

    if (edge == "short") ^ (aspect_ratio < 1.0):
        return edge_size, round(edge_size * aspect_ratio)
    else:
        return round(edge_size / aspect_ratio), edge_size


def extract_batch_size(x: torch.Tensor) -> int:
    verify_is_batched_image(x)
    return x.size()[0]


def extract_num_channels(x: torch.Tensor) -> int:
    verify_is_image(x)
    return x.size()[-3]


def extract_image_size(x: torch.Tensor) -> Tuple[int, int]:
    verify_is_image(x)
    return cast(Tuple[int, int], tuple(x.size()[-2:]))


def extract_edge_size(x: torch.Tensor, edge: str = "short") -> int:
    return image_to_edge_size(extract_image_size(x), edge=edge)


def extract_aspect_ratio(x: torch.Tensor) -> float:
    return calculate_aspect_ratio(extract_image_size(x))


def make_batched_image(x: torch.Tensor) -> torch.Tensor:
    verify_is_single_image(x)
    return x.unsqueeze(0)


def make_single_image(x: torch.Tensor) -> torch.Tensor:
    batch_size = extract_batch_size(x)
    if batch_size != 1:
        msg = "ADDME"  # FIXME
        raise RuntimeError(msg)
    return x.squeeze(0)


def force_image(fn: Callable) -> Callable:
    @functools.wraps(fn)
    def wrapper(x: torch.Tensor, *args: Any, **kwargs: Any) -> Any:
        verify_is_image(x)
        return fn(x, *args, **kwargs)

    return wrapper


def force_single_image(fn: Callable) -> Callable:
    @functools.wraps(fn)
    def wrapper(x: torch.Tensor, *args: Any, **kwargs: Any) -> Any:
        verify_is_image(x)
        is_batched = is_batched_image(x)
        if is_batched:
            x = make_single_image(x)

        x = fn(x, *args, **kwargs)

        if is_batched and is_image(x):
            x = make_batched_image(x)

        return x

    return wrapper


def force_batched_image(fn: Callable) -> Callable:
    @functools.wraps(fn)
    def wrapper(x: torch.Tensor, *args: Any, **kwargs: Any) -> Any:
        verify_is_image(x)
        is_single = is_single_image(x)
        if is_single:
            x = make_batched_image(x)

        x = fn(x, *args, **kwargs)

        if is_single and is_image(x):
            x = make_single_image(x)

        return x

    return wrapper



import warnings

from pystiche.misc import build_deprecation_message

from ._color import *
from ._crop import *
from ._misc import *
from ._motif import *
from ._resize import *

msg = build_deprecation_message(
    "Using functionality from pystiche.image.transforms.functional",
    "0.7.0",
    info="See https://github.com/pmeier/pystiche/issues/382 for details.",
)
warnings.warn(msg, UserWarning)



from typing import cast

import torch

from pystiche.image.utils import force_batched_image
from pystiche.meta import tensor_meta

from ._misc import transform_channels_linearly

__all__ = [
    "rgb_to_grayscale",
    "grayscale_to_fakegrayscale",
    "rgb_to_fakegrayscale",
    "grayscale_to_binary",
    "rgb_to_binary",
    "rgb_to_yuv",
    "yuv_to_rgb",
]


def rgb_to_grayscale(x: torch.Tensor) -> torch.Tensor:
    transformation_matrix = torch.tensor(((0.299, 0.587, 0.114),), **tensor_meta(x))
    return transform_channels_linearly(x, transformation_matrix)


@force_batched_image
def grayscale_to_fakegrayscale(x: torch.Tensor) -> torch.Tensor:
    return x.repeat(1, 3, 1, 1)


def rgb_to_fakegrayscale(x: torch.Tensor) -> torch.Tensor:
    return cast(torch.Tensor, grayscale_to_fakegrayscale(rgb_to_grayscale(x)))


def grayscale_to_binary(x: torch.Tensor) -> torch.Tensor:
    # TODO: add static / dynamic thresholding
    return torch.round(x)


def rgb_to_binary(x: torch.Tensor) -> torch.Tensor:
    return grayscale_to_binary(rgb_to_grayscale(x))


def rgb_to_yuv(x: torch.Tensor) -> torch.Tensor:
    transformation_matrix = torch.tensor(
        ((0.299, 0.587, 0.114), (-0.147, -0.289, 0.436), (0.615, -0.515, -0.100)),
        **tensor_meta(x),
    )
    return transform_channels_linearly(x, transformation_matrix)


def yuv_to_rgb(x: torch.Tensor) -> torch.Tensor:
    transformation_matrix = torch.tensor(
        ((1.000, 0.000, 1.140), (1.000, -0.395, -0.581), (1.000, 2.032, 0.000)),
        **tensor_meta(x),
    )
    return transform_channels_linearly(x, transformation_matrix)



from typing import List, Optional, Tuple, Union

import torch

from pystiche.ops import (
    ComparisonOperator,
    Operator,
    OperatorContainer,
    RegularizationOperator,
)

from .multi_op import MultiOperatorLoss

__all__ = ["PerceptualLoss", "GuidedPerceptualLoss"]


class _PerceptualLoss(MultiOperatorLoss):
    content_loss: Union[ComparisonOperator, OperatorContainer]
    style_loss: Union[ComparisonOperator, OperatorContainer]
    regularization: Optional[RegularizationOperator]

    def __init__(
        self,
        content_loss: Union[ComparisonOperator, OperatorContainer],
        style_loss: Union[ComparisonOperator, OperatorContainer],
        regularization: Optional[RegularizationOperator] = None,
        trim: bool = True,
    ) -> None:
        named_ops: List[Tuple[str, Operator]] = [
            ("content_loss", content_loss),
            ("style_loss", style_loss),
        ]

        if regularization is not None:
            named_ops.append(("regularization", regularization))

        super().__init__(named_ops, trim=trim)

    def set_content_image(self, image: torch.Tensor) -> None:
        r"""Set the content image.

        Args:
            image: Content image.
        """
        self.content_loss.set_target_image(image)


class PerceptualLoss(_PerceptualLoss):
    r"""Perceptual loss comprising content and style loss as well as optionally a
    regularization.

    Args:
        content_loss: Content loss.
        style_loss: Style loss.
        regularization: Optional regularization.
        trim: If ``True``, all :class:`~pystiche.enc.MultiLayerEncoder` s associated
            with ``content_loss``, ``style_loss``, or ``regularization`` will be
            :meth:`~pystiche.enc.MultiLayerEncoder.trim` med. Defaults to ``True``.
    """

    def set_style_image(self, image: torch.Tensor) -> None:
        r"""Set the style image.

        Args:
            image: Style image.
        """
        self.style_loss.set_target_image(image)


class GuidedPerceptualLoss(_PerceptualLoss):
    r"""Guided perceptual loss comprising content and guided style loss as well as
    optionally a regularization.

    Args:
        content_loss: Content loss.
        style_loss: Guided style loss.
        regularization: Optional regularization.
        trim: If ``True``, all :class:`~pystiche.enc.MultiLayerEncoder` s associated
            with ``content_loss``, ``style_loss``, or ``regularization`` will be
            :meth:`~pystiche.enc.MultiLayerEncoder.trim` med. Defaults to ``True``.
    """

    def set_style_guide(
        self, region: str, guide: torch.Tensor, recalc_repr: bool = True
    ) -> None:
        r"""Set the style guide for the specified region.

        Args:
            region: Region.
            guide: Style guide.
            recalc_repr: If ``True``, recalculates the style representation. See
                :meth:`pystiche.ops.ComparisonOperator.set_target_guide` for details.
                Defaults to ``True``.
        """
        getattr(self.style_loss, region).set_target_guide(
            guide, recalc_repr=recalc_repr
        )

    def set_style_image(self, region: str, image: torch.Tensor) -> None:
        r"""Set the style image for the specified region.

        Args:
            region: Region.
            image: Style image.
        """
        getattr(self.style_loss, region).set_target_image(image)

    def set_content_guide(self, region: str, guide: torch.Tensor) -> None:
        r"""Set the content guide for the specified region.

        Args:
            region: Region.
            guide: Content guide.
        """
        getattr(self.style_loss, region).set_input_guide(guide)



import re
import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import torchvision
from torch import hub, nn

from pystiche.misc import build_deprecation_message

from .utils import ModelMultiLayerEncoder
from .utils import select_url as _select_url

__all__ = [
    "VGGMultiLayerEncoder",
    "vgg11_multi_layer_encoder",
    "vgg11_bn_multi_layer_encoder",
    "vgg13_multi_layer_encoder",
    "vgg13_bn_multi_layer_encoder",
    "vgg16_multi_layer_encoder",
    "vgg16_bn_multi_layer_encoder",
    "vgg19_multi_layer_encoder",
    "vgg19_bn_multi_layer_encoder",
]

ARCH_PATTERN = re.compile(r"^vgg(?P<num_layers>(11|13|16|19)+)(?P<batch_norm>_bn)?$")

NUM_LAYERS_TO_CONFIGURATION = {
    11: "A",
    13: "B",
    16: "D",
    19: "E",
}


def _parse_arch(arch: str) -> Tuple[int, bool]:
    match = ARCH_PATTERN.match(arch)
    if match is None:
        raise ValueError(
            f"Unknown arch '{arch}'. It has to match 'vgg_(11|13|16|19)(_bn)?'"
        )

    num_layers = int(match.group("num_layers"))
    batch_norm = match.group("batch_norm") is not None

    return num_layers, batch_norm


def _make_description(arch: str, multi_layer_encoder: bool) -> str:
    num_layers, batch_norm = _parse_arch(arch)

    if multi_layer_encoder:
        short = (
            f"Multi-layer encoder based on :class:`~torchvision.models.VGG` "
            f"{num_layers}{' with batch normalization' if batch_norm else ''}."
        )
    else:
        short = f":class:`~torchvision.models.VGG` {num_layers} model"

    long = (
        f"    The :class:`~torchvision.models.VGG` architecture was introduced by "
        f"Krizhevsky, Sutskever, and Hinton in :cite:`KSH2012`. VGG{num_layers} "
        f"corresponds to configuration ``{NUM_LAYERS_TO_CONFIGURATION[num_layers]}`` "
        f"in the paper."
    )
    return "\n".join((short, "", long))


def _make_vgg_docstring(arch: str) -> str:
    description = _make_description(arch, multi_layer_encoder=False)
    args = r"""Args:
    pretrained: If ``True``, loads weights from training on
        :class:`~torchvision.models.ImageNet`. Defaults to ``False``.
    framework: Framework that was used to train the model. Can be one of
        ``"torch"`` (default) or ``"caffe"``.
        .. note::

            The weights for ``"caffe"`` were generated by Karen Simonyan and
            Andrew Zisserman. See https://download.pystiche.org/models/LICENSE for
            details.
    progress: If ``True``, displays a progress bar to STDERR during download of
        pretrained weights. Defaults to ``True``.
    num_classes: Size of the output layer. Defaults to ``1000``.
        .. note::

            Pretrained weights are only available for ``num_classes == 1000``.
    """
    return "\n".join((description, "", args))


def select_url(arch: str, framework: str) -> str:
    def format(key: Tuple[str, str]) -> str:
        arch, framework = key
        return "\n".join((f"arch={arch}", f"framework={framework}"))

    return _select_url(MODEL_URLS, (arch, framework), format=format)


def _vgg_loader(arch: str) -> Callable[..., torchvision.models.VGG]:
    loader = cast(
        Callable[..., torchvision.models.VGG], getattr(torchvision.models, arch)
    )

    def vgg(
        pretrained: bool = False,
        framework: str = "torch",
        progress: bool = True,
        num_classes: int = 1000,
    ) -> torchvision.models.VGG:
        if pretrained and num_classes != 1000:
            raise RuntimeError

        model = loader(pretrained=False, num_classes=num_classes)

        if not pretrained:
            return model

        state_dict = hub.load_state_dict_from_url(
            select_url(arch, framework), progress=progress, check_hash=True,
        )
        model.load_state_dict(state_dict)
        return model

    vgg.__doc__ = _make_vgg_docstring(arch)

    return vgg


TORCH_MODEL_URLS = torchvision.models.vgg.model_urls
ARCHS = tuple(TORCH_MODEL_URLS.keys())
MODEL_URLS = {(arch, "torch"): TORCH_MODEL_URLS[arch] for arch in ARCHS}
MODEL_URLS.update(
    {
        ("vgg16", "caffe"): "https://download.pystiche.org/models/vgg16-781be684.pth",
        ("vgg19", "caffe"): "https://download.pystiche.org/models/vgg19-74e45263.pth",
    }
)
MODELS = {arch: _vgg_loader(arch) for arch in ARCHS}


class VGGMultiLayerEncoder(ModelMultiLayerEncoder):
    r"""Multi-layer encoder based on :class:`~torchvision.models.VGG`.

    The :class:`~torchvision.models.VGG` architecture was introduced by  Krizhevsky,
    Sutskever, and Hinton in :cite:`KSH2012`

    Args:
        arch: :class:`~torchvision.models.VGG` architecture. Has to match
            ``"vgg(11|13|16|19)(_bn)?"``.
        pretrained: If ``True``, loads builtin weights. Defaults to ``True``.
        framework: Name of the framework that was used to train the builtin weights.
            Defaults to ``"torch"``.
        kwargs: Optional arguments of :class:`ModelMultiLayerEncoder` .

    Raises:
        RuntimeError: If ``pretrained is True`` and no weights are available for the
            combination of ``arch`` and ``framework``.
    """

    def __init__(self, arch: str, weights: Optional[str] = None, **kwargs: Any) -> None:
        if weights is not None:
            msg = build_deprecation_message(
                "The parameter weights", "0.6.0", info="It was renamed to framework"
            )
            warnings.warn(msg, UserWarning)
            kwargs["framework"] = weights

        _parse_arch(arch)
        self.arch = arch
        super().__init__(**kwargs)

    def state_dict_url(self, framework: str) -> str:
        return select_url(self.arch, framework)

    def collect_modules(
        self, inplace: bool
    ) -> Tuple[List[Tuple[str, nn.Module]], Dict[str, str]]:
        model = MODELS[self.arch](pretrained=False)

        modules = []
        state_dict_key_map = {}
        block = depth = 1
        for idx, module in model.features.named_children():
            if isinstance(module, nn.Conv2d):
                name = f"conv{block}_{depth}"
            elif isinstance(module, nn.BatchNorm2d):
                name = f"bn{block}_{depth}"
            elif isinstance(module, nn.ReLU):
                module = nn.ReLU(inplace=inplace)
                name = f"relu{block}_{depth}"
                # each ReLU layer increases the depth of the current block
                depth += 1
            else:  # isinstance(module, nn.MaxPool2d):
                name = f"pool{block}"
                # each pooling layer marks the end of the current block
                block += 1
                depth = 1

            modules.append((name, module))
            state_dict_key_map.update(
                {
                    f"features.{idx}.{key}": f"{name}.{key}"
                    for key in module.state_dict().keys()
                }
            )

        return modules, state_dict_key_map

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()

        dct["arch"] = self.arch
        dct.move_to_end("arch", last=False)  # type: ignore[attr-defined]

        return dct


def _make_vgg_multi_layer_encoder_docstring(arch: str) -> str:
    description = _make_description(arch, multi_layer_encoder=True)
    args = r"""    Args:
        kwargs: Optional arguments of :class:`VGGMultiLayerEncoder` .
    """
    return "\n".join((description, "", args))


def vgg11_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg11", **kwargs)


def vgg11_bn_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg11_bn", **kwargs)


def vgg13_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg13", **kwargs)


def vgg13_bn_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg13_bn", **kwargs)


def vgg16_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg16", **kwargs)


def vgg16_bn_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg16_bn", **kwargs)


def vgg19_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg19", **kwargs)


def vgg19_bn_multi_layer_encoder(**kwargs: Any) -> VGGMultiLayerEncoder:
    return VGGMultiLayerEncoder("vgg19_bn", **kwargs)


for arch in ARCHS:
    name = f"{arch}_multi_layer_encoder"
    doc = _make_vgg_multi_layer_encoder_docstring(arch)
    globals()[name].__doc__ = doc



from typing import Any, Dict, Optional, Tuple, Union

from PIL import Image

import torch

from pystiche.image import io

from .core import Transform

__all__ = ["ImportFromPIL", "ExportToPIL"]


class ImportFromPIL(Transform):
    def __init__(
        self, device: Union[torch.device, str] = "cpu", make_batched: bool = True
    ):
        super().__init__()
        self.device = device
        self.make_batched = make_batched

    def forward(self, x: Image.Image) -> torch.Tensor:
        return io.import_from_pil(x, self.device, make_batched=self.make_batched)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        device = str(self.device)
        if device != "cpu":
            dct["device"] = device
        dct["make_batched"] = self.make_batched
        return dct


class ExportToPIL(Transform):
    def __init__(self, mode: Optional[str] = None):
        super().__init__()
        self.mode = mode

    def forward(self, x: torch.Tensor) -> Union[Image.Image, Tuple[Image.Image, ...]]:
        return io.export_to_pil(x, mode=self.mode)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        if self.mode is not None:
            dct["mode"] = self.mode
        return dct



import os

__all__ = ["home"]


def home() -> str:
    r"""Local directory to save downloaded images and guides. Defaults to
    ``~/.cache/pystiche`` but can be overwritten with the ``PYSTICHE_HOME`` environment
    variable.
    """
    return os.getenv(
        "PYSTICHE_HOME", os.path.expanduser(os.path.join("~", ".cache", "pystiche"))
    )



from types import TracebackType
from typing import Iterator, Sequence, Tuple, Type

import torch
from torch import nn

import pystiche
from pystiche import enc, ops
from pystiche.misc import suppress_warnings

__all__ = ["MLEHandler", "MultiOperatorLoss"]


class MLEHandler(pystiche.ComplexObject):
    def __init__(self, criterion: nn.Module) -> None:
        self.multi_layer_encoders = {
            encoding_op.encoder.multi_layer_encoder
            for encoding_op in criterion.modules()
            if isinstance(encoding_op, ops.EncodingOperator)
            and isinstance(encoding_op.encoder, enc.SingleLayerEncoder)
        }

    def encode(self, input_image: torch.Tensor) -> None:
        for encoder in self.multi_layer_encoders:
            encoder.encode(input_image)

    def empty_storage(self) -> None:
        for mle in self.multi_layer_encoders:
            mle.empty_storage()

    def trim(self) -> None:
        for mle in self.multi_layer_encoders:
            mle.trim()

    def __call__(self, input_image: torch.Tensor) -> "MLEHandler":
        with suppress_warnings(FutureWarning):
            for mle in self.multi_layer_encoders:
                mle.encode(input_image)
        return self

    def __enter__(self) -> None:
        pass

    def __exit__(
        self, exc_type: Type[Exception], exc_val: Exception, exc_tb: TracebackType
    ) -> None:
        for encoder in self.multi_layer_encoders:
            encoder.empty_storage()

    def _named_children(self) -> Iterator[Tuple[str, enc.MultiLayerEncoder]]:
        return ((str(idx), mle) for idx, mle in enumerate(self.multi_layer_encoders))


class MultiOperatorLoss(pystiche.Module):
    r"""Generic loss for multiple :class:`~pystiche.ops.Operator` s. If called with an
    image it is passed to all immediate children operators and the
    results are returned as a :class:`pystiche.LossDict`. For that each
    :class:`pystiche.enc.MultiLayerEncoder` is only hit once, even if it is associated
    with multiple of the called operators.

    Args:
        named_ops: Named children operators.
        trim: If ``True``, all :class:`~pystiche.enc.MultiLayerEncoder` s associated
            with ``content_loss``, ``style_loss``, or ``regularization`` will be
            :meth:`~pystiche.enc.MultiLayerEncoder.trim` med. Defaults to ``True``.
    """

    def __init__(
        self, named_ops: Sequence[Tuple[str, ops.Operator]], trim: bool = True
    ) -> None:
        super().__init__(named_children=named_ops)
        self._mle_handler = MLEHandler(self)

        if trim:
            self._mle_handler.trim()

    def named_operators(
        self, recurse: bool = False
    ) -> Iterator[Tuple[str, ops.Operator]]:
        iterator = self.named_modules() if recurse else self.named_children()
        for name, child in iterator:
            if isinstance(child, ops.Operator):
                yield name, child

    def operators(self, recurse: bool = False) -> Iterator[ops.Operator]:
        for _, op in self.named_operators(recurse=recurse):
            yield op

    def forward(self, input_image: torch.Tensor) -> pystiche.LossDict:
        with self._mle_handler(input_image):
            return pystiche.LossDict(
                [(name, op(input_image)) for name, op in self.named_children()]
            )



from abc import ABC
from collections import OrderedDict
from copy import copy
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    SupportsFloat,
    Tuple,
    Union,
    cast,
)

import torch

from pystiche.meta import is_scalar_tensor
from pystiche.misc import build_complex_obj_repr, build_fmtstr

__all__ = ["ComplexObject", "LossDict", "TensorKey"]


class ComplexObject(ABC):
    r"""Object with a complex representation. See
    :func:`pystiche.misc.build_complex_obj_repr` for details.
    """
    _STR_INDENT = 2

    def _properties(self) -> Dict[str, Any]:
        r"""
        Returns:
            Internal properties.

        .. note::

            If subclassed, this method should integrate the new properties in the
            properties of the superclass.
        """
        return OrderedDict()

    def extra_properties(self) -> Dict[str, Any]:
        r"""
        Returns:
            Extra properties.
        """
        return OrderedDict()

    def properties(self) -> Dict[str, Any]:
        r"""
        Returns:
            Internal and extra properties.
        """
        dct = self._properties()
        dct.update(self.extra_properties())
        return dct

    def _named_children(self) -> Iterator[Tuple[str, Any]]:
        r"""
        Yields:
            Internal named children.

        .. note::

            If subclassed, this method should yield the named children of the
            superclass alongside yielding the new named children.
        """
        return iter(())

    def extra_named_children(self) -> Iterator[Tuple[str, Any]]:
        r"""
        Yields:
            Extra named children.
        """
        return iter(())

    def named_children(self) -> Iterator[Tuple[str, Any]]:
        r"""
        Yields:
            Internal and extra named children.
        """
        yield from self._named_children()
        yield from self.extra_named_children()

    def _build_repr(
        self,
        name: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
        named_children: Optional[Sequence[Tuple[str, Any]]] = None,
    ) -> str:
        if name is None:
            name = self.__class__.__name__

        if properties is None:
            properties = self.properties()

        if named_children is None:
            named_children = tuple(self.named_children())

        return build_complex_obj_repr(
            name,
            properties=properties,
            named_children=named_children,
            num_indent=self._STR_INDENT,
        )

    def __repr__(self) -> str:
        return self._build_repr()


class LossDict(OrderedDict):
    r"""Hierarchic dictionary of scalar :class:`torch.Tensor` losses. Levels are
    seperated by ``"."`` in the names.

    Args:
        losses: Optional named losses.
    """

    def __init__(
        self, losses: Sequence[Tuple[str, Union[torch.Tensor, "LossDict"]]] = (),
    ) -> None:
        super().__init__()
        for name, loss in losses:
            self[name] = loss

    def __setitem__(self, name: str, loss: Union[torch.Tensor, "LossDict"]) -> None:
        r"""Add a named loss to the entries.

        Args:
            name: Name of the loss.
            loss: If :class:`torch.Tensor`, it has to be scalar. If :class:`LossDict`,
                it is unpacked and the entries are added as level below of ``name``.

        Raises:
            TypeError: If loss is :class:`torch.Tensor` but isn't scalar.
        """
        if isinstance(loss, torch.Tensor):
            if not is_scalar_tensor(loss):
                msg = "loss is a torch.Tensor but is not scalar."
                raise TypeError(msg)
            super().__setitem__(name, loss)
        elif isinstance(loss, LossDict):
            for child_name, child_loss in loss.items():
                super().__setitem__(f"{name}.{child_name}", child_loss)
        else:
            msg = (  # type: ignore[unreachable]
                f"loss can be a scalar torch.Tensor or a pystiche.LossDict, but got "
                f"a {type(loss)} instead."
            )
            raise TypeError(msg)

    def aggregate(self, max_depth: int) -> Union[torch.Tensor, "LossDict"]:
        r"""Aggregate all entries up to a given maximum depth.

        Args:
            max_depth: If ``0`` returns sum of all entries as scalar
                :class:`torch.Tensor`.
        """

        def sum_partial_losses(partial_losses: Iterable[torch.Tensor]) -> torch.Tensor:
            return cast(torch.Tensor, sum(partial_losses))

        if max_depth == 0:
            return sum_partial_losses(self.values())

        splits = [name.split(".") for name in self.keys()]
        if not any([len(split) >= max_depth for split in splits]):
            return copy(self)

        agg_names = [".".join(split[:max_depth]) for split in splits]
        key_map = dict(zip(self.keys(), agg_names))
        agg_losses: Dict[str, List[torch.Tensor]] = {
            name: [] for name in set(agg_names)
        }
        for name, loss in self.items():
            agg_losses[key_map[name]].append(loss)

        return LossDict(
            [(name, sum_partial_losses(agg_losses[name])) for name in agg_names]
        )

    def total(self) -> torch.Tensor:
        r"""
        Returns:
            Sum of all entries as scalar tensor.
        """
        return cast(torch.Tensor, self.aggregate(0))

    def backward(self, *args: Any, **kwargs: Any) -> None:
        r"""Computes the gradient of all entries with respect to the graph leaves. See
        :meth:`torch.Tensor.backward` for details.
        """
        self.total().backward(*args, **kwargs)

    def item(self) -> float:
        r"""
        Returns:
            The sum of all entries as standard Python number.
        """
        return self.total().item()

    def __float__(self) -> float:
        return self.item()

    def __mul__(self, other: SupportsFloat) -> "LossDict":
        r"""Multiplies all entries with a scalar.

        Args:
            other: Scalar multiplier.
        """
        other = float(other)
        return LossDict([(name, loss * other) for name, loss in self.items()])

    def __str__(self) -> str:
        key_fmtstr = build_fmtstr(
            field_len=max([len(key) for key in self.keys()]), type="s"
        )
        value_fmtstr = build_fmtstr(precision=3, type="e")
        fmtstr = f"{key_fmtstr}: {value_fmtstr}"
        return "\n".join([fmtstr.format(key, value) for key, value in self.items()])


class TensorKey:
    def __init__(self, x: torch.Tensor, precision: int = 4) -> None:
        x = x.detach()
        self._key = (*self._extract_meta(x), *self._calculate_stats(x, precision))

    @staticmethod
    def _extract_meta(x: torch.Tensor) -> Tuple[Hashable, ...]:
        return (x.device, x.dtype, x.size())

    @staticmethod
    def _calculate_stats(x: torch.Tensor, precision: int) -> List[str]:
        stat_fns: Tuple[Callable[[torch.Tensor], torch.Tensor], ...] = (
            torch.min,
            torch.max,
            torch.norm,
        )
        return [f"{stat_fn(x).item():.{precision}e}" for stat_fn in stat_fns]

    @property
    def key(self) -> Tuple[Hashable, ...]:
        return self._key

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, torch.Tensor):
            other = TensorKey(other)

        return self.key == other.key if isinstance(other, TensorKey) else False

    def __hash__(self) -> int:
        return hash(self.key)

    def __repr__(self) -> str:
        return str(self.key)



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, Sequence

__all__ = [
    "License",
    "UnknownLicense",
    "NoLicense",
    "PublicDomainLicense",
    "ExpiredCopyrightLicense",
    "PixabayLicense",
    "CreativeCommonsLicense",
]


class License(ABC):
    @abstractmethod
    def __repr__(self) -> str:
        pass


class UnknownLicense(License):
    def __repr__(self) -> str:
        return (
            "The license for this image is unknown. "
            "Proceed to work with this image at your own risk."
        )


class NoLicense(License):
    def __repr__(self) -> str:
        return (
            "There is no license available for this image. "
            "Proceed to work with this image at your own risk."
        )


class PublicDomainLicense(License):
    def __repr__(self) -> str:
        return (
            "The copyright holder released this image into the public domain. This "
            "grants anyone the right to use this image for any purpose, without any "
            "conditions, unless such conditions are required by law."
        )


class ExpiredCopyrightLicense(PublicDomainLicense):
    def __init__(self, author_death_year: int):
        self.author_death_year = author_death_year

    def __repr__(self) -> str:
        years_since_author_death = str(datetime.now().year - self.author_death_year)
        return (
            f"This image is in the public domain in countries and areas, "
            f"where the copyright term is at most the authors life plus "
            f"{years_since_author_death} years. Check your local laws before "
            f"proceeding."
        )


class PixabayLicense(License):
    def __repr__(self) -> str:
        return "Pixabay (https://pixabay.com/service/license/)"


class CreativeCommonsLicense(License):
    TYPE_DICT = {
        "by": "Attribution",
        "sa": "ShareAlike",
        "nc": "NonCommercial",
        "nd": "NoDerivatives",
    }

    def __init__(
        self, types: Sequence[str], version: str, variant: Optional[str] = None,
    ) -> None:
        self.types = [type.lower() for type in types]
        self.version = version
        if variant is None:
            if version == "2.0":
                variant = "Generic"
            elif version == "3.0":
                variant = "Unported"
        self.variant = variant

    def __repr__(self) -> str:
        return self._create_license()

    def _create_license(self) -> str:
        long = "{0} {1}".format(
            "-".join([self.TYPE_DICT[type] for type in self.types]), self.version
        )
        if self.variant:
            long += f" {self.variant}"
        short = f"(CC {'-'.join([type.upper() for type in self.types])} {self.version})"
        url = "https://creativecommons.org/licenses/{0}/{1}".format(
            "-".join(self.types), self.version
        )
        return f"{long} {short} {url}"



from pystiche.misc import suppress_warnings

with suppress_warnings():
    from .color import *
    from .core import *
    from .crop import *
    from .io import *
    from .misc import *
    from .motif import *
    from .processing import *
    from .resize import *



from typing import Any, Dict, Optional, Tuple, Union, cast

import torch

from . import functional as F
from .core import Transform

__all__ = [
    "TransformMotifAffinely",
    "ShearMotif",
    "RotateMotif",
    "ScaleMotif",
    "TranslateMotif",
]


class TransformMotifAffinely(Transform):
    def __init__(
        self,
        shearing_angle: Optional[float] = None,
        clockwise_shearing: bool = False,
        shearing_center: Optional[Tuple[float, float]] = None,
        rotation_angle: Optional[float] = None,
        clockwise_rotation: bool = False,
        rotation_center: Optional[Tuple[float, float]] = None,
        scaling_factor: Optional[Union[float, Tuple[float, float]]] = None,
        scaling_center: Optional[Tuple[float, float]] = None,
        translation: Optional[Tuple[float, float]] = None,
        inverse_translation: bool = False,
        canvas: str = "same",
        interpolation_mode: str = "bilinear",
        padding_mode: str = "zeros",
    ) -> None:
        super().__init__()
        self.shearing_angle = shearing_angle
        self.clockwise_shearing = clockwise_shearing
        self.shearing_center = shearing_center
        self.rotation_angle = rotation_angle
        self.clockwise_rotation = clockwise_rotation
        self.rotation_center = rotation_center
        self.scaling_factor = scaling_factor
        self.scaling_center = scaling_center
        self.translation = translation
        self.inverse_translation = inverse_translation
        self.canvas = canvas
        self.interpolation_mode = interpolation_mode
        self.padding_mode = padding_mode

        if not self.has_affine_params:
            msg = (
                "TransformMotifAffinely was created without any affine parameter. At "
                "least one of shearing_angle, rotation_angle, scaling_factor, and "
                "translation has to be set."
            )
            raise RuntimeError(msg)

    @property
    def has_shearing(self) -> bool:
        return self.shearing_angle is not None

    @property
    def has_rotation(self) -> bool:
        return self.rotation_angle is not None

    @property
    def has_scaling(self) -> bool:
        return self.scaling_factor is not None

    @property
    def has_translation(self) -> bool:
        return self.translation is not None

    @property
    def has_affine_params(self) -> bool:
        return any(
            (
                self.has_rotation,
                self.has_shearing,
                self.has_scaling,
                self.has_translation,
            )
        )

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return cast(
            torch.Tensor,
            F.transform_motif_affinely(
                image,
                shearing_angle=self.shearing_angle,
                clockwise_shearing=self.clockwise_shearing,
                shearing_center=self.shearing_center,
                rotation_angle=self.rotation_angle,
                clockwise_rotation=self.clockwise_rotation,
                rotation_center=self.rotation_center,
                scaling_factor=self.scaling_factor,
                scaling_center=self.scaling_center,
                translation=self.translation,
                inverse_translation=self.inverse_translation,
                canvas=self.canvas,
                interpolation_mode=self.interpolation_mode,
                padding_mode=self.padding_mode,
            ),
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        if self.has_shearing:
            dct["shearing_angle"] = f"{self.shearing_angle}°"
            if self.clockwise_shearing:
                dct["clockwise_shearing"] = self.clockwise_shearing
            if self.shearing_center is not None:
                dct["shearing_center"] = self.shearing_center
        if self.has_rotation:
            dct["rotation_angle"] = f"{self.rotation_angle}°"
            if self.clockwise_rotation:
                dct["clockwise_rotation"] = self.clockwise_rotation
            if self.rotation_center is not None:
                dct["rotation_center"] = self.rotation_center
        if self.has_scaling:
            dct["scaling_factor"] = self.scaling_factor
            if self.scaling_center is not None:
                dct["scaling_center"] = self.scaling_center
        if self.has_translation:
            dct["translation"] = self.translation
            if self.inverse_translation:
                dct["inverse_translation"] = self.inverse_translation
        if self.canvas != "same":
            dct["canvas"] = self.canvas
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        if self.padding_mode != "zeros":
            dct["padding_mode"] = self.padding_mode
        return dct


class ShearMotif(Transform):
    def __init__(
        self,
        angle: float,
        clockwise: bool = False,
        center: Optional[Tuple[float, float]] = None,
        canvas: str = "same",
        interpolation_mode: str = "bilinear",
        padding_mode: str = "zeros",
    ) -> None:
        super().__init__()
        self.angle = angle
        self.clockwise = clockwise
        self.center = center
        self.canvas = canvas
        self.interpolation_mode = interpolation_mode
        self.padding_mode = padding_mode

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.shear_motif(
            image,
            self.angle,
            clockwise=self.clockwise,
            center=self.center,
            canvas=self.canvas,
            interpolation_mode=self.interpolation_mode,
            padding_mode=self.padding_mode,
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["angle"] = f"{self.angle}°"
        if self.clockwise:
            dct["clockwise"] = self.clockwise
        if self.center is not None:
            dct["center"] = self.center
        if self.canvas != "same":
            dct["canvas"] = self.canvas
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        if self.padding_mode != "zeros":
            dct["padding_mode"] = self.padding_mode
        return dct


class RotateMotif(Transform):
    def __init__(
        self,
        angle: float,
        clockwise: bool = False,
        center: Optional[Tuple[float, float]] = None,
        canvas: str = "same",
        interpolation_mode: str = "bilinear",
        padding_mode: str = "zeros",
    ) -> None:
        super().__init__()
        self.angle = angle
        self.clockwise = clockwise
        self.center = center
        self.canvas = canvas
        self.interpolation_mode = interpolation_mode
        self.padding_mode = padding_mode

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.rotate_motif(
            image,
            self.angle,
            clockwise=self.clockwise,
            center=self.center,
            canvas=self.canvas,
            interpolation_mode=self.interpolation_mode,
            padding_mode=self.padding_mode,
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["angle"] = f"{self.angle}°"
        if self.clockwise:
            dct["clockwise"] = self.clockwise
        if self.center is not None:
            dct["center"] = self.center
        if self.canvas != "same":
            dct["canvas"] = self.canvas
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        if self.padding_mode != "zeros":
            dct["padding_mode"] = self.padding_mode
        return dct


class ScaleMotif(Transform):
    def __init__(
        self,
        factor: Union[float, Tuple[float, float]],
        center: Optional[Tuple[float, float]] = None,
        canvas: str = "same",
        interpolation_mode: str = "bilinear",
        padding_mode: str = "zeros",
    ) -> None:
        super().__init__()
        self.factor = factor
        self.center = center
        self.canvas = canvas
        self.interpolation_mode = interpolation_mode
        self.padding_mode = padding_mode

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.scale_motif(
            image,
            self.factor,
            center=self.center,
            canvas=self.canvas,
            interpolation_mode=self.interpolation_mode,
            padding_mode=self.padding_mode,
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["factor"] = self.factor
        if self.center is not None:
            dct["center"] = self.center
        if self.canvas != "same":
            dct["canvas"] = self.canvas
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        if self.padding_mode != "zeros":
            dct["padding_mode"] = self.padding_mode
        return dct


class TranslateMotif(Transform):
    def __init__(
        self,
        translation: Tuple[float, float],
        inverse: bool = False,
        canvas: str = "same",
        interpolation_mode: str = "bilinear",
        padding_mode: str = "zeros",
    ):
        super().__init__()
        self.translation = translation
        self.inverse = inverse
        self.canvas = canvas
        self.interpolation_mode = interpolation_mode
        self.padding_mode = padding_mode

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.translate_motif(
            image,
            self.translation,
            inverse=self.inverse,
            canvas=self.canvas,
            interpolation_mode=self.interpolation_mode,
            padding_mode=self.padding_mode,
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["translation"] = self.translation
        if self.inverse:
            dct["inverse"] = self.inverse
        if self.canvas != "same":
            dct["canvas"] = self.canvas
        if self.interpolation_mode != "bilinear":
            dct["interpolation_mode"] = self.interpolation_mode
        if self.padding_mode != "zeros":
            dct["padding_mode"] = self.padding_mode
        return dct



import warnings
from typing import Any, Dict, List, Optional, Tuple

from torch import nn
from torchvision.models import alexnet
from torchvision.models.alexnet import model_urls as TORCH_MODEL_URLS

from pystiche.misc import build_deprecation_message

from .utils import ModelMultiLayerEncoder, select_url

__all__ = ["AlexNetMultiLayerEncoder", "alexnet_multi_layer_encoder"]


MODEL_URLS = {"torch": TORCH_MODEL_URLS["alexnet"]}


def _make_description() -> str:
    return r"""Multi-layer encoder based on :class:`~torchvision.models.AlexNet`.

    The :class:`~torchvision.models.AlexNet` architecture was introduced by
    Krizhevsky, Sutskever, and Hinton in :cite:`KSH2012`.
    """


def _make_docstring(body: str) -> str:
    return f"{_make_description()}\n{body}"


class AlexNetMultiLayerEncoder(ModelMultiLayerEncoder):
    __doc__ = _make_docstring(
        r"""    Args:
        pretrained: If ``True``, loads builtin weights. Defaults to ``True``.
        framework: Name of the framework that was used to train the builtin weights.
            Defaults to ``"torch"``.
        kwargs: Optional arguments of :class:`ModelMultiLayerEncoder` .

    Raises:
        RuntimeError: If ``pretrained`` and no weights are available for the
        ``framework``.
        """
    )

    def __init__(self, weights: Optional[str] = None, **kwargs: Any) -> None:
        if weights is not None:
            msg = build_deprecation_message(
                "The parameter weights", "0.6.0", info="It was renamed to framework"
            )
            warnings.warn(msg, UserWarning)
            kwargs["framework"] = weights

        super().__init__(**kwargs)

    def state_dict_url(self, framework: str) -> str:
        return select_url(MODEL_URLS, framework)

    def collect_modules(
        self, inplace: bool
    ) -> Tuple[List[Tuple[str, nn.Module]], Dict[str, str]]:
        model = alexnet(pretrained=False)

        modules = []
        state_dict_key_map = {}
        block = 1
        for idx, module in model.features.named_children():
            if isinstance(module, nn.Conv2d):
                name = f"conv{block}"
            elif isinstance(module, nn.ReLU):
                module = nn.ReLU(inplace=inplace)
                name = f"relu{block}"
                if block in (3, 4):
                    # in the third and forth block the ReLU layer marks the end of the
                    # block
                    block += 1
            else:  # isinstance(module, nn.MaxPool2d):
                name = f"pool{block}"
                # each pooling layer marks the end of the current block
                block += 1

            modules.append((name, module))
            state_dict_key_map.update(
                {
                    f"features.{idx}.{key}": f"{name}.{key}"
                    for key in module.state_dict().keys()
                }
            )

        return modules, state_dict_key_map


def alexnet_multi_layer_encoder(**kwargs: Any) -> AlexNetMultiLayerEncoder:
    return AlexNetMultiLayerEncoder(**kwargs)


alexnet_multi_layer_encoder.__doc__ = _make_docstring(
    r"""    Args:
        kwargs: Optional arguments of :class:`AlexNetMultiLayerEncoder` .
"""
)



from typing import Any, Dict, Tuple, Union, cast

import torch

from pystiche.image import extract_image_size

from . import functional as F
from .core import Transform
from .functional._crop import _parse_size

__all__ = [
    "Crop",
    "CenterCrop",
    "TopLeftCrop",
    "BottomLeftCrop",
    "TopRightCrop",
    "BottomRightCrop",
    "ValidRandomCrop",
]


class Crop(Transform):
    def __init__(
        self,
        origin: Tuple[int, int],
        size: Union[Tuple[int, int], int],
        vert_anchor: str = "top",
        horz_anchor: str = "left",
    ):
        super().__init__()
        self.origin = origin
        self.size = size
        self.vert_anchor = vert_anchor
        self.horz_anchor = horz_anchor

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return cast(
            torch.Tensor,
            F.crop(
                image,
                self.origin,
                self.size,
                vert_anchor=self.vert_anchor,
                horz_anchor=self.horz_anchor,
            ),
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["origin"] = self.origin
        dct["size"] = self.size
        if self.vert_anchor != "top" or self.horz_anchor != "left":
            dct["vert_anchor"] = self.vert_anchor
            dct["horz_anchor"] = self.horz_anchor
        return dct


class TopLeftCrop(Transform):
    def __init__(self, size: Union[Tuple[int, int], int]):
        super().__init__()
        self.size = size

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.top_left_crop(image, self.size)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["size"] = self.size
        return dct


class BottomLeftCrop(Transform):
    def __init__(self, size: Union[Tuple[int, int], int]):
        super().__init__()
        self.size = size

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.bottom_left_crop(image, self.size)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["size"] = self.size
        return dct


class TopRightCrop(Transform):
    def __init__(self, size: Union[Tuple[int, int], int]):
        super().__init__()
        self.size = size

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.top_right_crop(image, self.size)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["size"] = self.size
        return dct


class BottomRightCrop(Transform):
    def __init__(self, size: Union[Tuple[int, int], int]):
        super().__init__()
        self.size = size

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.bottom_right_crop(image, self.size)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["size"] = self.size
        return dct


class CenterCrop(Transform):
    def __init__(self, size: Union[Tuple[int, int], int]):
        super().__init__()
        self.size = size

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        return F.center_crop(image, self.size)

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["size"] = self.size
        return dct


class ValidRandomCrop(Transform):
    def __init__(self, size: Union[Tuple[int, int], int]):
        super().__init__()
        self.size = size

    @staticmethod
    def get_random_origin(
        image_size: Tuple[int, int], crop_size: Union[Tuple[int, int], int]
    ) -> Tuple[int, int]:
        image_height, image_width = image_size
        crop_height, crop_width = _parse_size(crop_size)

        def randint(range: int) -> int:
            if range < 0:
                msg = "The crop size has to be smaller or equal to the image size."
                raise RuntimeError(msg)
            return cast(int, torch.randint(range + 1, (), dtype=torch.long).item())

        vert_origin = randint(image_height - crop_height)
        horz_origin = randint(image_width - crop_width)
        return (vert_origin, horz_origin)

    def forward(self, image: torch.Tensor) -> torch.Tensor:
        origin = self.get_random_origin(extract_image_size(image), self.size)
        return cast(torch.Tensor, F.crop(image, origin, self.size))

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        dct["size"] = self.size
        return dct



from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, cast

import torch
from torch import hub, nn
from torch.nn.modules.module import _IncompatibleKeys

from ..multi_layer_encoder import MultiLayerEncoder
from ..preprocessing import get_preprocessor

__all__ = ["ModelMultiLayerEncoder", "select_url"]

T = TypeVar("T")


def select_url(
    urls: Dict[T, str], key: T, format: Optional[Callable[[T], str]] = None
) -> str:
    if format is None:
        format = str

    try:
        return urls[key]
    except KeyError as error:
        raise RuntimeError(f"No URL is available for\n\n{format(key)}") from error


class ModelMultiLayerEncoder(MultiLayerEncoder):
    r"""Multi-layer encoder based on a pre-defined model.

    Args:
        pretrained: If ``True``, loads builtin weights. Defaults to ``True``.
        framework: Name of the framework that was used to train the builtin weights.
            Defaults to ``"torch"``.
        internal_preprocessing: If ``True``, adds a preprocessing layer for the
            selected ``framework`` as first layer. Defaults to ``True``.
        allow_inplace: If ``True``, allows inplace operations to reduce the memory
            requirement during the forward pass. Defaults to ``False``.

            .. warning::
                After performing an inplace operation the encodings of the previous
                layer is no longer accessible. Only use this if you are sure that you
                do **not** need these encodings.
    """

    def __init__(
        self,
        pretrained: bool = True,
        framework: str = "torch",
        internal_preprocessing: bool = True,
        allow_inplace: bool = False,
    ) -> None:
        self.pretrained = pretrained
        self.framework = framework
        self.internal_preprocessing = internal_preprocessing
        self.allow_inplace = allow_inplace

        modules, self._state_dict_key_map = self.collect_modules(allow_inplace)
        if internal_preprocessing:
            modules.insert(0, ("preprocessing", get_preprocessor(framework)))

        super().__init__(modules)

        if pretrained:
            self.load_state_dict_from_url(framework)

    @abstractmethod
    def state_dict_url(self, framework: str) -> str:
        r"""Select URL of a downloadable ``state_dict``.

        Args:
            framework: Name of the framework that was used to train the weights.

        Raises:
            RuntimeError: If no ``state_dict`` is available.
        """
        pass

    @abstractmethod
    def collect_modules(
        self, inplace: bool
    ) -> Tuple[List[Tuple[str, nn.Module]], Dict[str, str]]:
        r"""Collect modules of a base model with more descriptive names.

        Args:
            inplace: If ``True``, when possible, modules should use inplace operations.

        Returns:
            List of name-module-pairs as well as a dictionary mapping the new, more
            descriptive names to the original ones.
        """
        pass

    def _map_state_dict_keys(
        self, state_dict: Dict[str, torch.Tensor]
    ) -> Tuple[Dict[str, torch.Tensor], List[str]]:
        remapped_state_dict = {}
        unexpected_keys = []
        for key, value in state_dict.items():
            if key in self._state_dict_key_map:
                remapped_state_dict[self._state_dict_key_map[key]] = value
            else:
                unexpected_keys.append(key)

        return remapped_state_dict, unexpected_keys

    def load_state_dict(
        self,
        state_dict: Dict[str, torch.Tensor],
        strict: bool = True,
        map_names: bool = True,
        framework: str = "unknown",
    ) -> _IncompatibleKeys:
        r"""Loads parameters and buffers from the ``state_dict``.

        Args:
            state_dict: State dictionary.
            strict: Enforce matching keys in ``state_dict`` and the internal states.
            map_names: If ``True``, maps the names names in ``state_dict`` of the
                underlying model to the more descriptive names generated by
                :meth:`collect_modules`. Defaults to ``True``.
            framework: Name of the framework that was used to train the weights in
                ``state_dict``. Defaults to ``"unknown"``.

                .. note::

                    This has no effect on the behavior, but makes the representation
                    of the :class:`ModelMultiLayerEncoder` more descriptive.

        Returns:
            Named tuple with ``missing_keys`` and ``unexpected_keys`` fields.

        .. seealso::

            :meth:`torch.nn.Module.load_state_dict`
        """
        if map_names:
            state_dict, unexpected_keys = self._map_state_dict_keys(state_dict)
        else:
            unexpected_keys = []

        keys = cast(
            _IncompatibleKeys,
            super().load_state_dict(state_dict, strict=strict),  # type: ignore[arg-type]
        )
        keys.unexpected_keys.extend(unexpected_keys)

        self.pretrained = True
        self.framework = framework

        return keys

    def load_state_dict_from_url(
        self,
        framework: str,
        strict: bool = True,
        map_names: bool = True,
        check_hash: bool = True,
        **kwargs: Any,
    ) -> None:
        r"""Downloads and loads parameters and buffers trained with ``framework``.

        Args:
            framework: Name of the framework that was used to train the weights of the
                ``state_dict``.
            strict: Enforce matching keys in ``state_dict`` and the internal states.
            map_names: If ``True``, maps the names names in ``state_dict`` of the
                underlying model to the more descriptive names generated by
                :meth:`collect_modules`. Defaults to ``True``.
            check_hash: If ``True``, checks if the hash postfix of the URL matches the
                SHA256 hash of the downloaded ``state_dict``. Defaults to ``True``.
            kwargs: Optional arguments for :meth:`torch.hub.load_state_dict_from_url` .

        .. seealso::

            - :meth:`state_dict_url`
            - :meth:`load_state_dict`
            - :meth:`torch.hub.load_state_dict_from_url`
        """
        url = self.state_dict_url(framework)
        state_dict = hub.load_state_dict_from_url(url, check_hash=check_hash, **kwargs)
        self.load_state_dict(
            state_dict, strict=strict, map_names=map_names, framework=framework
        )

    def _properties(self) -> Dict[str, Any]:
        dct = super()._properties()
        if not self.pretrained:
            dct["pretrained"] = False
        else:
            dct["framework"] = self.framework
        if not self.internal_preprocessing:
            dct["internal_preprocessing"] = self.internal_preprocessing
        if self.allow_inplace:
            dct["allow_inplace"] = self.allow_inplace
        return dct
